# -*- coding: utf-8 -*-
from __future__ import unicode_literals

try:
    basestring
except NameError:
    basestring = str

import platform

# Defaults to Windows canonical names (platform-specific overrides below)
canonical_names = {
    'escape': 'esc',
    'return': 'enter',
    'del': 'delete',
    'control': 'ctrl',

    'left arrow': 'left',
    'up arrow': 'up',
    'down arrow': 'down',
    'right arrow': 'right',

    ' ': 'space', # Prefer to spell out keys that would be hard to read.
    '\x1b': 'esc',
    '\x08': 'backspace',
    '\n': 'enter',
    '\t': 'tab',
    '\r': 'enter',

    'scrlk': 'scroll lock',
    'prtscn': 'print screen',
    'prnt scrn': 'print screen',
    'snapshot': 'print screen',
    'ins': 'insert',
    'pause break': 'pause',
    'ctrll lock': 'caps lock',
    'capslock': 'caps lock',
    'number lock': 'num lock',
    'numlock': 'num lock',
    'space bar': 'space',
    'spacebar': 'space',
    'linefeed': 'enter',
    'win': 'windows',

    # Mac keys
    'command': 'windows',
    'cmd': 'windows',
    'control': 'ctrl',
    'option': 'alt',

    'app': 'menu',
    'apps': 'menu',
    'application': 'menu',
    'applications': 'menu',

    'pagedown': 'page down',
    'pageup': 'page up',
    'pgdown': 'page down',
    'pgup': 'page up',

    'play/pause': 'play/pause media',

    'num multiply': '*',
    'num divide': '/',
    'num add': '+',
    'num plus': '+',
    'num minus': '-',
    'num sub': '-',
    'num enter': 'enter',
    'num 0': '0',
    'num 1': '1',
    'num 2': '2',
    'num 3': '3',
    'num 4': '4',
    'num 5': '5',
    'num 6': '6',
    'num 7': '7',
    'num 8': '8',
    'num 9': '9',

    'left win': 'left windows',
    'right win': 'right windows',
    'left control': 'left ctrl',
    'right control': 'right ctrl',
    'left menu': 'left alt', # Windows...
    'altgr': 'alt gr',

    # https://www.x.org/releases/X11R7.6/doc/libX11/Compose/en_US.UTF-8.html
    # https://svn.apache.org/repos/asf/xmlgraphics/commons/tags/commons-1_0/src/java/org/apache/xmlgraphics/fonts/Glyphs.java
    # Note this list has plenty of uppercase letters that are not being used
    # at the moment, as normalization forces names to be lowercase.
    "Aacute": "Á",
    "aacute": "á",
    "Aacutesmall": "",
    "abovedot": "˙",
    "Abreve": "Ă",
    "abreve": "ă",
    "Abreveacute": "Ắ",
    "abreveacute": "ắ",
    "Abrevebelowdot": "Ặ",
    "abrevebelowdot": "ặ",
    "Abrevegrave": "Ằ",
    "abrevegrave": "ằ",
    "Abrevehook": "Ẳ",
    "abrevehook": "ẳ",
    "Abrevetilde": "Ẵ",
    "abrevetilde": "ẵ",
    "Acircumflex": "Â",
    "acircumflex": "â",
    "Acircumflexacute": "Ấ",
    "acircumflexacute": "ấ",
    "Acircumflexbelowdot": "Ậ",
    "acircumflexbelowdot": "ậ",
    "Acircumflexgrave": "Ầ",
    "acircumflexgrave": "ầ",
    "Acircumflexhook": "Ẩ",
    "acircumflexhook": "ẩ",
    "Acircumflexsmall": "",
    "Acircumflextilde": "Ẫ",
    "acircumflextilde": "ẫ",
    "acute": "´",
    "Acute": "",
    "acutecomb": "́",
    "Acutesmall": "",
    "add": "+",
    "Adiaeresis": "Ä",
    "adiaeresis": "ä",
    "Adieresis": "Ä",
    "adieresis": "ä",
    "Adieresissmall": "",
    "ae": "æ",
    "AE": "Æ",
    "AEacute": "Ǽ",
    "aeacute": "ǽ",
    "AEsmall": "",
    "afii00208": "―",
    "afii10017": "А",
    "afii10018": "Б",
    "afii10019": "В",
    "afii10020": "Г",
    "afii10021": "Д",
    "afii10022": "Е",
    "afii10023": "Ё",
    "afii10024": "Ж",
    "afii10025": "З",
    "afii10026": "И",
    "afii10027": "Й",
    "afii10028": "К",
    "afii10029": "Л",
    "afii10030": "М",
    "afii10031": "Н",
    "afii10032": "О",
    "afii10033": "П",
    "afii10034": "Р",
    "afii10035": "С",
    "afii10036": "Т",
    "afii10037": "У",
    "afii10038": "Ф",
    "afii10039": "Х",
    "afii10040": "Ц",
    "afii10041": "Ч",
    "afii10042": "Ш",
    "afii10043": "Щ",
    "afii10044": "Ъ",
    "afii10045": "Ы",
    "afii10046": "Ь",
    "afii10047": "Э",
    "afii10048": "Ю",
    "afii10049": "Я",
    "afii10050": "Ґ",
    "afii10051": "Ђ",
    "afii10052": "Ѓ",
    "afii10053": "Є",
    "afii10054": "Ѕ",
    "afii10055": "І",
    "afii10056": "Ї",
    "afii10057": "Ј",
    "afii10058": "Љ",
    "afii10059": "Њ",
    "afii10060": "Ћ",
    "afii10061": "Ќ",
    "afii10062": "Ў",
    "afii10063": "",
    "afii10064": "",
    "afii10065": "а",
    "afii10066": "б",
    "afii10067": "в",
    "afii10068": "г",
    "afii10069": "д",
    "afii10070": "е",
    "afii10071": "ё",
    "afii10072": "ж",
    "afii10073": "з",
    "afii10074": "и",
    "afii10075": "й",
    "afii10076": "к",
    "afii10077": "л",
    "afii10078": "м",
    "afii10079": "н",
    "afii10080": "о",
    "afii10081": "п",
    "afii10082": "р",
    "afii10083": "с",
    "afii10084": "т",
    "afii10085": "у",
    "afii10086": "ф",
    "afii10087": "х",
    "afii10088": "ц",
    "afii10089": "ч",
    "afii10090": "ш",
    "afii10091": "щ",
    "afii10092": "ъ",
    "afii10093": "ы",
    "afii10094": "ь",
    "afii10095": "э",
    "afii10096": "ю",
    "afii10097": "я",
    "afii10098": "ґ",
    "afii10099": "ђ",
    "afii10100": "ѓ",
    "afii10101": "є",
    "afii10102": "ѕ",
    "afii10103": "і",
    "afii10104": "ї",
    "afii10105": "ј",
    "afii10106": "љ",
    "afii10107": "њ",
    "afii10108": "ћ",
    "afii10109": "ќ",
    "afii10110": "ў",
    "afii10145": "Џ",
    "afii10146": "Ѣ",
    "afii10147": "Ѳ",
    "afii10148": "Ѵ",
    "afii10192": "",
    "afii10193": "џ",
    "afii10194": "ѣ",
    "afii10195": "ѳ",
    "afii10196": "ѵ",
    "afii10831": "",
    "afii10832": "",
    "afii10846": "ә",
    "afii299": "‎",
    "afii300": "‏",
    "afii301": "‍",
    "afii57381": "٪",
    "afii57388": "،",
    "afii57392": "٠",
    "afii57393": "١",
    "afii57394": "٢",
    "afii57395": "٣",
    "afii57396": "٤",
    "afii57397": "٥",
    "afii57398": "٦",
    "afii57399": "٧",
    "afii57400": "٨",
    "afii57401": "٩",
    "afii57403": "؛",
    "afii57407": "؟",
    "afii57409": "ء",
    "afii57410": "آ",
    "afii57411": "أ",
    "afii57412": "ؤ",
    "afii57413": "إ",
    "afii57414": "ئ",
    "afii57415": "ا",
    "afii57416": "ب",
    "afii57417": "ة",
    "afii57418": "ت",
    "afii57419": "ث",
    "afii57420": "ج",
    "afii57421": "ح",
    "afii57422": "خ",
    "afii57423": "د",
    "afii57424": "ذ",
    "afii57425": "ر",
    "afii57426": "ز",
    "afii57427": "س",
    "afii57428": "ش",
    "afii57429": "ص",
    "afii57430": "ض",
    "afii57431": "ط",
    "afii57432": "ظ",
    "afii57433": "ع",
    "afii57434": "غ",
    "afii57440": "ـ",
    "afii57441": "ف",
    "afii57442": "ق",
    "afii57443": "ك",
    "afii57444": "ل",
    "afii57445": "م",
    "afii57446": "ن",
    "afii57448": "و",
    "afii57449": "ى",
    "afii57450": "ي",
    "afii57451": "ً",
    "afii57452": "ٌ",
    "afii57453": "ٍ",
    "afii57454": "َ",
    "afii57455": "ُ",
    "afii57456": "ِ",
    "afii57457": "ّ",
    "afii57458": "ْ",
    "afii57470": "ه",
    "afii57505": "ڤ",
    "afii57506": "پ",
    "afii57507": "چ",
    "afii57508": "ژ",
    "afii57509": "گ",
    "afii57511": "ٹ",
    "afii57512": "ڈ",
    "afii57513": "ڑ",
    "afii57514": "ں",
    "afii57519": "ے",
    "afii57534": "ە",
    "afii57636": "₪",
    "afii57645": "־",
    "afii57658": "׃",
    "afii57664": "א",
    "afii57665": "ב",
    "afii57666": "ג",
    "afii57667": "ד",
    "afii57668": "ה",
    "afii57669": "ו",
    "afii57670": "ז",
    "afii57671": "ח",
    "afii57672": "ט",
    "afii57673": "י",
    "afii57674": "ך",
    "afii57675": "כ",
    "afii57676": "ל",
    "afii57677": "ם",
    "afii57678": "מ",
    "afii57679": "ן",
    "afii57680": "נ",
    "afii57681": "ס",
    "afii57682": "ע",
    "afii57683": "ף",
    "afii57684": "פ",
    "afii57685": "ץ",
    "afii57686": "צ",
    "afii57687": "ק",
    "afii57688": "ר",
    "afii57689": "ש",
    "afii57690": "ת",
    "afii57694": "שׁ",
    "afii57695": "שׂ",
    "afii57700": "וֹ",
    "afii57705": "ײַ",
    "afii57716": "װ",
    "afii57717": "ױ",
    "afii57718": "ײ",
    "afii57723": "וּ",
    "afii57793": "ִ",
    "afii57794": "ֵ",
    "afii57795": "ֶ",
    "afii57796": "ֻ",
    "afii57797": "ָ",
    "afii57798": "ַ",
    "afii57799": "ְ",
    "afii57800": "ֲ",
    "afii57801": "ֱ",
    "afii57802": "ֳ",
    "afii57803": "ׂ",
    "afii57804": "ׁ",
    "afii57806": "ֹ",
    "afii57807": "ּ",
    "afii57839": "ֽ",
    "afii57841": "ֿ",
    "afii57842": "׀",
    "afii57929": "ʼ",
    "afii61248": "℅",
    "afii61289": "ℓ",
    "afii61352": "№",
    "afii61573": "‬",
    "afii61574": "‭",
    "afii61575": "‮",
    "afii61664": "‌",
    "afii63167": "٭",
    "afii64937": "ʽ",
    "Agrave": "À",
    "agrave": "à",
    "Agravesmall": "",
    "agudo": "´",
    "aleph": "ℵ",
    "Alpha": "Α",
    "alpha": "α",
    "Alphatonos": "Ά",
    "alphatonos": "ά",
    "Amacron": "Ā",
    "amacron": "ā",
    "ampersand": "&",
    "ampersandsmall": "",
    "angle": "∠",
    "angleleft": "〈",
    "angleright": "〉",
    "anoteleia": "·",
    "Aogonek": "Ą",
    "aogonek": "ą",
    "apostrophe": "'",
    "approxequal": "≈",
    "Aring": "Å",
    "aring": "å",
    "Aringacute": "Ǻ",
    "aringacute": "ǻ",
    "Aringsmall": "",
    "arrowboth": "↔",
    "arrowdblboth": "⇔",
    "arrowdbldown": "⇓",
    "arrowdblleft": "⇐",
    "arrowdblright": "⇒",
    "arrowdblup": "⇑",
    "arrowdown": "↓",
    "arrowhorizex": "",
    "arrowleft": "←",
    "arrowright": "→",
    "arrowup": "↑",
    "arrowupdn": "↕",
    "arrowupdnbse": "↨",
    "arrowvertex": "",
    "asciicircum": "^",
    "asciitilde": "~",
    "Asmall": "",
    "asterisk": "*",
    "asteriskmath": "∗",
    "asuperior": "",
    "at": "@",
    "Atilde": "Ã",
    "atilde": "ã",
    "Atildesmall": "",
    "backslash": "\\",
    "bar": "|",
    "Beta": "Β",
    "beta": "β",
    "block": "█",
    "braceex": "",
    "braceleft": "{",
    "braceleftbt": "",
    "braceleftmid": "",
    "bracelefttp": "",
    "braceright": "}",
    "bracerightbt": "",
    "bracerightmid": "",
    "bracerighttp": "",
    "bracketleft": "[",
    "bracketleftbt": "",
    "bracketleftex": "",
    "bracketlefttp": "",
    "bracketright": "]",
    "bracketrightbt": "",
    "bracketrightex": "",
    "bracketrighttp": "",
    "breve": "˘",
    "Brevesmall": "",
    "brokenbar": "¦",
    "Bsmall": "",
    "bsuperior": "",
    "bullet": "•",
    "Cacute": "Ć",
    "cacute": "ć",
    "caron": "ˇ",
    "Caron": "",
    "Caronsmall": "",
    "carriagereturn": "↵",
    "Ccaron": "Č",
    "ccaron": "č",
    "Ccedilla": "Ç",
    "ccedilla": "ç",
    "Ccedillasmall": "",
    "Ccircumflex": "Ĉ",
    "ccircumflex": "ĉ",
    "Cdotaccent": "Ċ",
    "cdotaccent": "ċ",
    "cedilla": "¸",
    "Cedillasmall": "",
    "cent": "¢",
    "centinferior": "",
    "centoldstyle": "",
    "centsuperior": "",
    "Chi": "Χ",
    "chi": "χ",
    "circle": "○",
    "circlemultiply": "⊗",
    "circleplus": "⊕",
    "circumflex": "^",
    "circumflex": "ˆ",
    "Circumflexsmall": "",
    "club": "♣",
    "colon": ":",
    "colonmonetary": "₡",
    "ColonSign": "₡",
    "comma": ",",
    "commaaccent": "",
    "commainferior": "",
    "commasuperior": "",
    "congruent": "≅",
    "copyright": "©",
    "copyrightsans": "",
    "copyrightserif": "",
    "CruzeiroSign": "₢",
    "Csmall": "",
    "currency": "¤",
    "cyrBreve": "",
    "cyrbreve": "",
    "cyrFlex": "",
    "cyrflex": "",
    "dagger": "†",
    "daggerdbl": "‡",
    "dblGrave": "",
    "dblgrave": "",
    "Dcaron": "Ď",
    "dcaron": "ď",
    "Dcroat": "Đ",
    "dcroat": "đ",
    "degree": "°",
    "Delta": "Δ",
    "delta": "δ",
    "diaeresis": "¨",
    "diamond": "♦",
    "dieresis": "¨",
    "Dieresis": "",
    "DieresisAcute": "",
    "dieresisacute": "",
    "DieresisGrave": "",
    "dieresisgrave": "",
    "Dieresissmall": "",
    "dieresistonos": "΅",
    "divide": "/",
    "divide": "÷",
    "division": "÷",
    "dkshade": "▓",
    "dnblock": "▄",
    "dollar": "$",
    "dollarinferior": "",
    "dollaroldstyle": "",
    "dollarsuperior": "",
    "dong": "₫",
    "DongSign": "₫",
    "dot": ".",
    "dotaccent": "˙",
    "Dotaccentsmall": "",
    "dotbelowcomb": "̣",
    "dotlessi": "ı",
    "dotlessj": "",
    "dotmath": "⋅",
    "Dsmall": "",
    "dstroke": "đ",
    "Dstroke": "Đ",
    "dsuperior": "",
    "Eacute": "É",
    "eacute": "é",
    "Eacutesmall": "",
    "Ebreve": "Ĕ",
    "ebreve": "ĕ",
    "Ecaron": "Ě",
    "ecaron": "ě",
    "Ecircumflex": "Ê",
    "ecircumflex": "ê",
    "Ecircumflexacute": "Ế",
    "ecircumflexacute": "ế",
    "Ecircumflexbelowdot": "Ệ",
    "ecircumflexbelowdot": "ệ",
    "Ecircumflexgrave": "Ề",
    "ecircumflexgrave": "ề",
    "Ecircumflexhook": "Ể",
    "ecircumflexhook": "ể",
    "Ecircumflexsmall": "",
    "Ecircumflextilde": "Ễ",
    "ecircumflextilde": "ễ",
    "EcuSign": "₠",
    "Ediaeresis": "Ë",
    "ediaeresis": "ë",
    "Edieresis": "Ë",
    "edieresis": "ë",
    "Edieresissmall": "",
    "Edotaccent": "Ė",
    "edotaccent": "ė",
    "Egrave": "È",
    "egrave": "è",
    "Egravesmall": "",
    "eight": "8",
    "eightinferior": "₈",
    "eightoldstyle": "",
    "eightsubscript": "₈",
    "eightsuperior": "⁸",
    "element": "∈",
    "ellipsis": "…",
    "Emacron": "Ē",
    "emacron": "ē",
    "emdash": "—",
    "emptyset": "∅",
    "endash": "–",
    "enfilledcircbullet": "•",
    "Eng": "Ŋ",
    "eng": "ŋ",
    "Eogonek": "Ę",
    "eogonek": "ę",
    "Epsilon": "Ε",
    "epsilon": "ε",
    "Epsilontonos": "Έ",
    "epsilontonos": "έ",
    "equal": "=",
    "equivalence": "≡",
    "Esmall": "",
    "estimated": "℮",
    "esuperior": "",
    "Eta": "Η",
    "eta": "η",
    "Etatonos": "Ή",
    "etatonos": "ή",
    "ETH": "Ð",
    "eth": "ð",
    "Eth": "Ð",
    "Ethsmall": "",
    "euro": "€",
    "Euro": "€",
    "EuroSign": "€",
    "exclam": "!",
    "exclamdbl": "‼",
    "exclamdown": "¡",
    "exclamdownsmall": "",
    "exclamsmall": "",
    "existential": "∃",
    "female": "♀",
    "ff": "ﬀ",
    "ffi": "ﬃ",
    "ffl": "ﬄ",
    "FFrancSign": "₣",
    "fi": "ﬁ",
    "figuredash": "‒",
    "filledbox": "■",
    "filledrect": "▬",
    "five": "5",
    "fiveeighths": "⅝",
    "fiveinferior": "₅",
    "fiveoldstyle": "",
    "fivesubscript": "₅",
    "fivesuperior": "⁵",
    "fl": "ﬂ",
    "florin": "ƒ",
    "four": "4",
    "fourinferior": "₄",
    "fouroldstyle": "",
    "foursubscript": "₄",
    "foursuperior": "⁴",
    "fraction": "∕",
    "franc": "₣",
    "Fsmall": "",
    "function": "ƒ",
    "Gamma": "Γ",
    "gamma": "γ",
    "Gbreve": "Ğ",
    "gbreve": "ğ",
    "Gcaron": "Ǧ",
    "gcaron": "ǧ",
    "Gcircumflex": "Ĝ",
    "gcircumflex": "ĝ",
    "Gcommaaccent": "Ģ",
    "gcommaaccent": "ģ",
    "Gdotaccent": "Ġ",
    "gdotaccent": "ġ",
    "germandbls": "ß",
    "gradient": "∇",
    "grave": "`",
    "Grave": "",
    "gravecomb": "̀",
    "Gravesmall": "",
    "greater": ">",
    "greaterequal": "≥",
    "Gsmall": "",
    "guillemotleft": "«",
    "guillemotright": "»",
    "guilsinglleft": "‹",
    "guilsinglright": "›",
    "H18533": "●",
    "H18543": "▪",
    "H18551": "▫",
    "H22073": "□",
    "hash": "#",
    "hashtag": "#",
    "Hbar": "Ħ",
    "hbar": "ħ",
    "Hcircumflex": "Ĥ",
    "hcircumflex": "ĥ",
    "heart": "♥",
    "hookabovecomb": "̉",
    "house": "⌂",
    "Hsmall": "",
    "hungarumlaut": "˝",
    "Hungarumlaut": "",
    "Hungarumlautsmall": "",
    "hyphen": "­",
    "hypheninferior": "",
    "hyphensuperior": "",
    "Iacute": "Í",
    "iacute": "í",
    "Iacutesmall": "",
    "Ibreve": "Ĭ",
    "ibreve": "ĭ",
    "Icircumflex": "Î",
    "icircumflex": "î",
    "Icircumflexsmall": "",
    "Idiaeresis": "Ï",
    "idiaeresis": "ï",
    "Idieresis": "Ï",
    "idieresis": "ï",
    "Idieresissmall": "",
    "Idotaccent": "İ",
    "Ifraktur": "ℑ",
    "Igrave": "Ì",
    "igrave": "ì",
    "Igravesmall": "",
    "IJ": "Ĳ",
    "ij": "ĳ",
    "Imacron": "Ī",
    "imacron": "ī",
    "infinity": "∞",
    "integral": "∫",
    "integralbt": "⌡",
    "integralex": "",
    "integraltp": "⌠",
    "intersection": "∩",
    "invbullet": "◘",
    "invcircle": "◙",
    "invsmileface": "☻",
    "Iogonek": "Į",
    "iogonek": "į",
    "Iota": "Ι",
    "iota": "ι",
    "Iotadieresis": "Ϊ",
    "iotadieresis": "ϊ",
    "iotadieresistonos": "ΐ",
    "Iotatonos": "Ί",
    "iotatonos": "ί",
    "Ismall": "",
    "isuperior": "",
    "Itilde": "Ĩ",
    "itilde": "ĩ",
    "Jcircumflex": "Ĵ",
    "jcircumflex": "ĵ",
    "Jsmall": "",
    "Kappa": "Κ",
    "kappa": "κ",
    "Kcommaaccent": "Ķ",
    "kcommaaccent": "ķ",
    "kgreenlandic": "ĸ",
    "Ksmall": "",
    "Lacute": "Ĺ",
    "lacute": "ĺ",
    "Lambda": "Λ",
    "lambda": "λ",
    "Lcaron": "Ľ",
    "lcaron": "ľ",
    "Lcommaaccent": "Ļ",
    "lcommaaccent": "ļ",
    "Ldot": "Ŀ",
    "ldot": "ŀ",
    "less": "<",
    "lessequal": "≤",
    "lfblock": "▌",
    "lira": "₤",
    "LiraSign": "₤",
    "LL": "",
    "ll": "",
    "logicaland": "∧",
    "logicalnot": "¬",
    "logicalor": "∨",
    "longs": "ſ",
    "lozenge": "◊",
    "Lslash": "Ł",
    "lslash": "ł",
    "Lslashsmall": "",
    "Lsmall": "",
    "lsuperior": "",
    "ltshade": "░",
    "macron": "¯",
    "macron": "ˉ",
    "Macron": "",
    "Macronsmall": "",
    "male": "♂",
    "masculine": "º",
    "MillSign": "₥",
    "minplus": "+",
    "minus": "-",
    "minus": "−",
    "minute": "′",
    "Msmall": "",
    "msuperior": "",
    "mu": "µ",
    "Mu": "Μ",
    "mu": "μ",
    "multiply": "*",
    "multiply": "×",
    "musicalnote": "♪",
    "musicalnotedbl": "♫",
    "Nacute": "Ń",
    "nacute": "ń",
    "NairaSign": "₦",
    "napostrophe": "ŉ",
    "Ncaron": "Ň",
    "ncaron": "ň",
    "Ncommaaccent": "Ņ",
    "ncommaaccent": "ņ",
    "NewSheqelSign": "₪",
    "nine": "9",
    "nineinferior": "₉",
    "nineoldstyle": "",
    "ninesubscript": "₉",
    "ninesuperior": "⁹",
    "nobreakspace": " ",
    "notelement": "∉",
    "notequal": "≠",
    "notsign": "¬",
    "notsubset": "⊄",
    "Nsmall": "",
    "nsuperior": "ⁿ",
    "Ntilde": "Ñ",
    "ntilde": "ñ",
    "Ntildesmall": "",
    "Nu": "Ν",
    "nu": "ν",
    "numbersign": "#",
    "numerosign": "№",
    "Oacute": "Ó",
    "oacute": "ó",
    "Oacutesmall": "",
    "Obreve": "Ŏ",
    "obreve": "ŏ",
    "Ocircumflex": "Ô",
    "ocircumflex": "ô",
    "Ocircumflexacute": "Ố",
    "ocircumflexacute": "ố",
    "Ocircumflexbelowdot": "Ộ",
    "ocircumflexbelowdot": "ộ",
    "Ocircumflexgrave": "Ồ",
    "ocircumflexgrave": "ồ",
    "Ocircumflexhook": "Ổ",
    "ocircumflexhook": "ổ",
    "Ocircumflexsmall": "",
    "Ocircumflextilde": "Ỗ",
    "ocircumflextilde": "ỗ",
    "Odiaeresis": "Ö",
    "odiaeresis": "ö",
    "Odieresis": "Ö",
    "odieresis": "ö",
    "Odieresissmall": "",
    "oe": "œ",
    "OE": "Œ",
    "OEsmall": "",
    "ogonek": "˛",
    "Ogoneksmall": "",
    "Ograve": "Ò",
    "ograve": "ò",
    "Ogravesmall": "",
    "Ohorn": "Ơ",
    "ohorn": "ơ",
    "Ohornacute": "Ớ",
    "ohornacute": "ớ",
    "Ohornbelowdot": "Ợ",
    "ohornbelowdot": "ợ",
    "Ohorngrave": "Ờ",
    "ohorngrave": "ờ",
    "Ohornhook": "Ở",
    "ohornhook": "ở",
    "Ohorntilde": "Ỡ",
    "ohorntilde": "ỡ",
    "Ohungarumlaut": "Ő",
    "ohungarumlaut": "ő",
    "Omacron": "Ō",
    "omacron": "ō",
    "Omega": "Ω",
    "omega": "ω",
    "omega1": "ϖ",
    "Omegatonos": "Ώ",
    "omegatonos": "ώ",
    "Omicron": "Ο",
    "omicron": "ο",
    "Omicrontonos": "Ό",
    "omicrontonos": "ό",
    "one": "1",
    "onedotenleader": "․",
    "oneeighth": "⅛",
    "onefitted": "",
    "onehalf": "½",
    "oneinferior": "₁",
    "oneoldstyle": "",
    "onequarter": "¼",
    "onesubscript": "₁",
    "onesuperior": "¹",
    "onethird": "⅓",
    "openbullet": "◦",
    "ordfeminine": "ª",
    "ordmasculine": "º",
    "orthogonal": "∟",
    "Oslash": "Ø",
    "oslash": "ø",
    "Oslashacute": "Ǿ",
    "oslashacute": "ǿ",
    "Oslashsmall": "",
    "Osmall": "",
    "osuperior": "",
    "Otilde": "Õ",
    "otilde": "õ",
    "Otildesmall": "",
    "paragraph": "¶",
    "parenleft": "(",
    "parenleftbt": "",
    "parenleftex": "",
    "parenleftinferior": "₍",
    "parenleftsuperior": "⁽",
    "parenlefttp": "",
    "parenright": ")",
    "parenrightbt": "",
    "parenrightex": "",
    "parenrightinferior": "₎",
    "parenrightsuperior": "⁾",
    "parenrighttp": "",
    "partialdiff": "∂",
    "percent": "%",
    "period": ".",
    "periodcentered": "·",
    "periodcentered": "∙",
    "periodinferior": "",
    "periodsuperior": "",
    "perpendicular": "⊥",
    "perthousand": "‰",
    "peseta": "₧",
    "PesetaSign": "₧",
    "Phi": "Φ",
    "phi": "φ",
    "phi1": "ϕ",
    "Pi": "Π",
    "pi": "π",
    "plus": "+",
    "plusminus": "±",
    "pound": "£",
    "prescription": "℞",
    "product": "∏",
    "propersubset": "⊂",
    "propersuperset": "⊃",
    "proportional": "∝",
    "Psi": "Ψ",
    "psi": "ψ",
    "Psmall": "",
    "Qsmall": "",
    "question": "?",
    "questiondown": "¿",
    "questiondownsmall": "",
    "questionsmall": "",
    "quotedbl": "\"",
    "quotedblbase": "„",
    "quotedblleft": "“",
    "quotedblright": "”",
    "quoteleft": "‘",
    "quotereversed": "‛",
    "quoteright": "’",
    "quotesinglbase": "‚",
    "quotesingle": "'",
    "Racute": "Ŕ",
    "racute": "ŕ",
    "radical": "√",
    "radicalex": "",
    "Rcaron": "Ř",
    "rcaron": "ř",
    "Rcommaaccent": "Ŗ",
    "rcommaaccent": "ŗ",
    "reflexsubset": "⊆",
    "reflexsuperset": "⊇",
    "registered": "®",
    "registersans": "",
    "registerserif": "",
    "revlogicalnot": "⌐",
    "Rfraktur": "ℜ",
    "Rho": "Ρ",
    "rho": "ρ",
    "ring": "˚",
    "Ringsmall": "",
    "Rsmall": "",
    "rsuperior": "",
    "rtblock": "▐",
    "RupeeSign": "₨",
    "rupiah": "",
    "Sacute": "Ś",
    "sacute": "ś",
    "Scaron": "Š",
    "scaron": "š",
    "Scaronsmall": "",
    "Scedilla": "",
    "scedilla": "",
    "Scircumflex": "Ŝ",
    "scircumflex": "ŝ",
    "Scommaaccent": "Ș",
    "scommaaccent": "ș",
    "second": "″",
    "section": "§",
    "semicolon": ";",
    "seven": "7",
    "seveneighths": "⅞",
    "seveninferior": "₇",
    "sevenoldstyle": "",
    "sevensubscript": "₇",
    "sevensuperior": "⁷",
    "SF010000": "┌",
    "SF020000": "└",
    "SF030000": "┐",
    "SF040000": "┘",
    "SF050000": "┼",
    "SF060000": "┬",
    "SF070000": "┴",
    "SF080000": "├",
    "SF090000": "┤",
    "SF100000": "─",
    "SF110000": "│",
    "SF190000": "╡",
    "SF200000": "╢",
    "SF210000": "╖",
    "SF220000": "╕",
    "SF230000": "╣",
    "SF240000": "║",
    "SF250000": "╗",
    "SF260000": "╝",
    "SF270000": "╜",
    "SF280000": "╛",
    "SF360000": "╞",
    "SF370000": "╟",
    "SF380000": "╚",
    "SF390000": "╔",
    "SF400000": "╩",
    "SF410000": "╦",
    "SF420000": "╠",
    "SF430000": "═",
    "SF440000": "╬",
    "SF450000": "╧",
    "SF460000": "╨",
    "SF470000": "╤",
    "SF480000": "╥",
    "SF490000": "╙",
    "SF500000": "╘",
    "SF510000": "╒",
    "SF520000": "╓",
    "SF530000": "╫",
    "SF540000": "╪",
    "shade": "▒",
    "Sigma": "Σ",
    "sigma": "σ",
    "sigma1": "ς",
    "similar": "∼",
    "similarequal": "≃",
    "six": "6",
    "sixinferior": "₆",
    "sixoldstyle": "",
    "sixsubscript": "₆",
    "sixsuperior": "⁶",
    "slash": "/",
    "smileface": "☺",
    "spade": "♠",
    "ssharp": "§",
    "ssharp": "ß",
    "Ssharp": "ẞ",
    "Ssmall": "",
    "ssuperior": "",
    "sterling": "£",
    "subtract": "-",
    "suchthat": "∋",
    "summation": "∑",
    "sun": "☼",
    "Tau": "Τ",
    "tau": "τ",
    "Tbar": "Ŧ",
    "tbar": "ŧ",
    "Tcaron": "Ť",
    "tcaron": "ť",
    "Tcommaaccent": "Ț",
    "tcommaaccent": "ț",
    "Thai_baht": "฿",
    "therefore": "∴",
    "Theta": "Θ",
    "theta": "θ",
    "theta1": "ϑ",
    "THORN": "Þ",
    "thorn": "þ",
    "Thorn": "Þ",
    "Thornsmall": "",
    "three": "3",
    "threeeighths": "⅜",
    "threeinferior": "₃",
    "threeoldstyle": "",
    "threequarters": "¾",
    "threequartersemdash": "",
    "threesubscript": "₃",
    "threesuperior": "³",
    "til": "~",
    "tilde": "~",
    "tilde": "˜",
    "tildecomb": "̃",
    "Tildesmall": "",
    "tonos": "΄",
    "trademark": "™",
    "trademarksans": "",
    "trademarkserif": "",
    "triagdn": "▼",
    "triaglf": "◄",
    "triagrt": "►",
    "triagup": "▲",
    "Tsmall": "",
    "tsuperior": "",
    "two": "2",
    "twodotenleader": "‥",
    "twoinferior": "₂",
    "twooldstyle": "",
    "twosubscript": "₂",
    "twosuperior": "²",
    "twothirds": "⅔",
    "Uacute": "Ú",
    "uacute": "ú",
    "Uacutesmall": "",
    "Ubreve": "Ŭ",
    "ubreve": "ŭ",
    "Ucircumflex": "Û",
    "ucircumflex": "û",
    "Ucircumflexsmall": "",
    "Udiaeresis": "Ü",
    "udiaeresis": "ü",
    "Udieresis": "Ü",
    "udieresis": "ü",
    "Udieresissmall": "",
    "Ugrave": "Ù",
    "ugrave": "ù",
    "Ugravesmall": "",
    "Uhorn": "Ư",
    "uhorn": "ư",
    "Uhornacute": "Ứ",
    "uhornacute": "ứ",
    "Uhornbelowdot": "Ự",
    "uhornbelowdot": "ự",
    "Uhorngrave": "Ừ",
    "uhorngrave": "ừ",
    "Uhornhook": "Ử",
    "uhornhook": "ử",
    "Uhorntilde": "Ữ",
    "uhorntilde": "ữ",
    "Uhungarumlaut": "Ű",
    "uhungarumlaut": "ű",
    "Umacron": "Ū",
    "umacron": "ū",
    "underscore": "_",
    "underscoredbl": "‗",
    "union": "∪",
    "universal": "∀",
    "Uogonek": "Ų",
    "uogonek": "ų",
    "upblock": "▀",
    "Upsilon": "Υ",
    "upsilon": "υ",
    "Upsilon1": "ϒ",
    "Upsilondieresis": "Ϋ",
    "upsilondieresis": "ϋ",
    "upsilondieresistonos": "ΰ",
    "Upsilontonos": "Ύ",
    "upsilontonos": "ύ",
    "Uring": "Ů",
    "uring": "ů",
    "Usmall": "",
    "Utilde": "Ũ",
    "utilde": "ũ",
    "Vsmall": "",
    "Wacute": "Ẃ",
    "wacute": "ẃ",
    "Wcircumflex": "Ŵ",
    "wcircumflex": "ŵ",
    "Wdieresis": "Ẅ",
    "wdieresis": "ẅ",
    "weierstrass": "℘",
    "Wgrave": "Ẁ",
    "wgrave": "ẁ",
    "WonSign": "₩",
    "Wsmall": "",
    "Xi": "Ξ",
    "xi": "ξ",
    "Xsmall": "",
    "Yacute": "Ý",
    "yacute": "ý",
    "Yacutesmall": "",
    "Ycircumflex": "Ŷ",
    "ycircumflex": "ŷ",
    "ydiaeresis": "ÿ",
    "Ydieresis": "Ÿ",
    "ydieresis": "ÿ",
    "Ydieresissmall": "",
    "yen": "¥",
    "Ygrave": "Ỳ",
    "ygrave": "ỳ",
    "Ysmall": "",
    "Zacute": "Ź",
    "zacute": "ź",
    "Zcaron": "Ž",
    "zcaron": "ž",
    "Zcaronsmall": "",
    "Zdotaccent": "Ż",
    "zdotaccent": "ż",
    "zero": "0",
    "zeroinferior": "₀",
    "zerooldstyle": "",
    "zerosubscript": "₀",
    "zerosuperior": "⁰",
    "zeta": "ζ",
    "Zeta": "Ζ",
    "Zsmall": "",
}
sided_modifiers = {'ctrl', 'alt', 'shift', 'windows'}
all_modifiers = {'alt', 'alt gr', 'ctrl', 'shift', 'windows'} | set('left ' + n for n in sided_modifiers) | set('right ' + n for n in sided_modifiers)

# Platform-specific canonical overrides

if platform.system() == 'Darwin':
    canonical_names.update({
        "command": "command",
        "windows": "command",
        "cmd": "command",
        "win": "command",
        "backspace": "delete",
        'alt gr': 'alt' # Issue #117
    })
    all_modifiers = {'alt', 'ctrl', 'shift', 'windows'}
if platform.system() == 'Linux':
    canonical_names.update({
        "select": "end",
        "find": "home",
        'next': 'page down',
        'prior': 'page up',
    })

def normalize_name(name):
    """
    Given a key name (e.g. "LEFT CONTROL"), clean up the string and convert to
    the canonical representation (e.g. "left ctrl") if one is known.
    """
    if not name or not isinstance(name, basestring):
        raise ValueError('Can only normalize non-empty string names. Unexpected '+ repr(name))

    if len(name) > 1:
        name = name.lower()
    if name != '_' and '_' in name:
        name = name.replace('_', ' ')

    return canonical_names.get(name, name)



# !/usr/bin/python
# coding=utf-8
import json

from pythontk.File import getFile


class Json():
	'''
	'''
	@classmethod
	def setJsonFile(cls, file):
		'''Set the current json filepath.

		:Parameters:
			file (str): The filepath to a json file. If a file doesn't exist, it will be created.
		'''
		cls._jsonFile = file
		getFile(cls._jsonFile) #will create the file if it does not exist.


	@classmethod
	def getJsonFile(cls):
		'''Get the current json filepath.

		:Return:
			(str)
		'''
		try:
			return cls._jsonFile
		except AttributeError as error:
			return ''


	@classmethod
	def setJson(cls, key, value, file=None):
		'''
		:Parameters:
			key () = Set the json key.
			value () = Set the json value for the given key.
			file (str): Temporarily set the filepath to a json file.
				If no file is given, the previously set file will be used 
				if one was set.

		:Example: setJson('hdr_map_visibility', state)
		'''
		if not file:
			file = cls.getJsonFile()

		assert file, '{} in setJson\n\t# Error: Operation requires a json file to be specified. #'.format(__file__)
		assert isinstance(file, str), '{} in setJson\n\t# Error:	Incorrect datatype: {} #'.format(__file__, type(file).__name__)

		try:
			with open(file, 'r') as f:
				dct = json.loads(f.read())
				dct[key] = value
		except json.decoder.JSONDecodeError as error:
			dct={}
			dct[key] = value

		with open(file, 'w') as f:
			f.write(json.dumps(dct))


	@classmethod
	def getJson(cls, key, file=None):
		'''
		:Parameters:
			key () = Set the json key.
			value () = Set the json value for the given key.
			file (str): Temporarily set the filepath to a json file.
					If no file is given, the previously set file will 
					be used if one was set.
		:Return:
			(str)

		:Example: getJson('hdr_map_visibility') #returns: state
		'''
		if not file:
			file = cls.getJsonFile()

		assert file, '{} in setJson\n\t# Error: Operation requires a json file to be specified. #'.format(__file__)
		assert isinstance(file, str), '{} in setJson\n\t# Error:	Incorrect datatype: {} #'.format(__file__, type(file).__name__)

		try:
			with open(file, 'r') as f:
				return json.loads(f.read())[key]

		except KeyError as error:
			# print ('# Error: {}: getJson: KeyError: {}'.format(__file__, error))
			pass
		except FileNotFoundError as error:
			# print ('# Error: {}: getJson: FileNotFoundError: {}'.format(__file__, error))
			pass
		except json.decoder.JSONDecodeError as error:
			print ('{} in getJson\n\t# Error: JSONDecodeError: {} #'.format(__file__, error))

# --------------------------------------------------------------------------------------------









# --------------------------------------------------------------------------------------------

def __getattr__(attr:str):
	"""Searches for an attribute in this module's classes and returns it.

	:Parameters:
		attr (str): The name of the attribute to search for.
	
	:Return:
		(obj) The found attribute.

	:Raises:
		AttributeError: If the given attribute is not found in any of the classes in the module.
	"""
	try:
		return getattr(Json, attr)

	except AttributeError as error:
		raise AttributeError(f"Module '{__name__}' has no attribute '{attr}'")

# --------------------------------------------------------------------------------------------

if __name__=='__main__':
	pass

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# Deprecated ------------------------------------


# PyAudio : Python Bindings for PortAudio.
#
# Copyright (c) 2006 Hubert Pham
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
PyAudio provides Python bindings for PortAudio, the cross-platform
audio I/O library. With PyAudio, you can easily use Python to play and
record audio on a variety of platforms.

.. include:: ../sphinx/examples.rst

Overview
--------

**Classes**
  :py:class:`PyAudio`, :py:class:`PyAudio.Stream`

.. only:: pamac

   **Host Specific Classes**
     :py:class:`PaMacCoreStreamInfo`

**Stream Conversion Convenience Functions**
  :py:func:`get_sample_size`, :py:func:`get_format_from_width`

**PortAudio version**
  :py:func:`get_portaudio_version`, :py:func:`get_portaudio_version_text`

.. |PaSampleFormat| replace:: :ref:`PortAudio Sample Format <PaSampleFormat>`
.. _PaSampleFormat:

**Portaudio Sample Formats**
  :py:data:`paFloat32`, :py:data:`paInt32`, :py:data:`paInt24`,
  :py:data:`paInt16`, :py:data:`paInt8`, :py:data:`paUInt8`,
  :py:data:`paCustomFormat`

.. |PaHostAPI| replace:: :ref:`PortAudio Host API <PaHostAPI>`
.. _PaHostAPI:

**PortAudio Host APIs**
  :py:data:`paInDevelopment`, :py:data:`paDirectSound`, :py:data:`paMME`,
  :py:data:`paASIO`, :py:data:`paSoundManager`, :py:data:`paCoreAudio`,
  :py:data:`paOSS`, :py:data:`paALSA`, :py:data:`paAL`, :py:data:`paBeOS`,
  :py:data:`paWDMKS`, :py:data:`paJACK`, :py:data:`paWASAPI`,
  :py:data:`paNoDevice`

.. |PaErrorCode| replace:: :ref:`PortAudio Error Code <PaErrorCode>`
.. _PaErrorCode:

**PortAudio Error Codes**
  :py:data:`paNoError`, :py:data:`paNotInitialized`,
  :py:data:`paUnanticipatedHostError`, :py:data:`paInvalidChannelCount`,
  :py:data:`paInvalidSampleRate`, :py:data:`paInvalidDevice`,
  :py:data:`paInvalidFlag`, :py:data:`paSampleFormatNotSupported`,
  :py:data:`paBadIODeviceCombination`, :py:data:`paInsufficientMemory`,
  :py:data:`paBufferTooBig`, :py:data:`paBufferTooSmall`,
  :py:data:`paNullCallback`, :py:data:`paBadStreamPtr`,
  :py:data:`paTimedOut`, :py:data:`paInternalError`,
  :py:data:`paDeviceUnavailable`,
  :py:data:`paIncompatibleHostApiSpecificStreamInfo`,
  :py:data:`paStreamIsStopped`, :py:data:`paStreamIsNotStopped`,
  :py:data:`paInputOverflowed`, :py:data:`paOutputUnderflowed`,
  :py:data:`paHostApiNotFound`, :py:data:`paInvalidHostApi`,
  :py:data:`paCanNotReadFromACallbackStream`,
  :py:data:`paCanNotWriteToACallbackStream`,
  :py:data:`paCanNotReadFromAnOutputOnlyStream`,
  :py:data:`paCanNotWriteToAnInputOnlyStream`,
  :py:data:`paIncompatibleStreamHostApi`

.. |PaCallbackReturnCodes| replace:: :ref:`PortAudio Callback Return Code <PaCallbackReturnCodes>`
.. _PaCallbackReturnCodes:

**PortAudio Callback Return Codes**
  :py:data:`paContinue`, :py:data:`paComplete`, :py:data:`paAbort`

.. |PaCallbackFlags| replace:: :ref:`PortAutio Callback Flag <PaCallbackFlags>`
.. _PaCallbackFlags:

**PortAudio Callback Flags**
  :py:data:`paInputUnderflow`, :py:data:`paInputOverflow`,
  :py:data:`paOutputUnderflow`, :py:data:`paOutputOverflow`,
  :py:data:`paPrimingOutput`
"""

__author__ = "Hubert Pham"
__version__ = "0.2.13"
__docformat__ = "restructuredtext en"

import locale
import warnings

try:
    import pyaudio._portaudio as pa
except ImportError:
    print("Could not import the PyAudio C module 'pyaudio._portaudio'.")
    raise


# PaSampleFormat Sample Formats

paFloat32 = pa.paFloat32  #: 32 bit float
paInt32 = pa.paInt32  #: 32 bit int
paInt24 = pa.paInt24  #: 24 bit int
paInt16 = pa.paInt16  #: 16 bit int
paInt8 = pa.paInt8  #: 8 bit int
paUInt8 = pa.paUInt8  #: 8 bit unsigned int
paCustomFormat = pa.paCustomFormat  #: a custom data format

# HostAPI TypeId

paInDevelopment = pa.paInDevelopment  #: Still in development
paDirectSound = pa.paDirectSound  #: DirectSound (Windows only)
paMME = pa.paMME  #: Multimedia Extension (Windows only)
paASIO = pa.paASIO  #: Steinberg Audio Stream Input/Output
paSoundManager = pa.paSoundManager  #: SoundManager (OSX only)
paCoreAudio = pa.paCoreAudio  #: CoreAudio (OSX only)
paOSS = pa.paOSS  #: Open Sound System (Linux only)
paALSA = pa.paALSA  #: Advanced Linux Sound Architecture (Linux only)
paAL = pa.paAL  #: Open Audio Library
paBeOS = pa.paBeOS  #: BeOS Sound System
paWDMKS = pa.paWDMKS  #: Windows Driver Model (Windows only)
paJACK = pa.paJACK  #: JACK Audio Connection Kit
paWASAPI = pa.paWASAPI  #: Windows Vista Audio stack architecture
paNoDevice = pa.paNoDevice  #: Not actually an audio device

# PortAudio Error Codes

paNoError = pa.paNoError
paNotInitialized = pa.paNotInitialized
paUnanticipatedHostError = pa.paUnanticipatedHostError
paInvalidChannelCount = pa.paInvalidChannelCount
paInvalidSampleRate = pa.paInvalidSampleRate
paInvalidDevice = pa.paInvalidDevice
paInvalidFlag = pa.paInvalidFlag
paSampleFormatNotSupported = pa.paSampleFormatNotSupported
paBadIODeviceCombination = pa.paBadIODeviceCombination
paInsufficientMemory = pa.paInsufficientMemory
paBufferTooBig = pa.paBufferTooBig
paBufferTooSmall = pa.paBufferTooSmall
paNullCallback = pa.paNullCallback
paBadStreamPtr = pa.paBadStreamPtr
paTimedOut = pa.paTimedOut
paInternalError = pa.paInternalError
paDeviceUnavailable = pa.paDeviceUnavailable
paIncompatibleHostApiSpecificStreamInfo = (
    pa.paIncompatibleHostApiSpecificStreamInfo)
paStreamIsStopped = pa.paStreamIsStopped
paStreamIsNotStopped = pa.paStreamIsNotStopped
paInputOverflowed = pa.paInputOverflowed
paOutputUnderflowed = pa.paOutputUnderflowed
paHostApiNotFound = pa.paHostApiNotFound
paInvalidHostApi = pa.paInvalidHostApi
paCanNotReadFromACallbackStream = pa.paCanNotReadFromACallbackStream
paCanNotWriteToACallbackStream = pa.paCanNotWriteToACallbackStream
paCanNotReadFromAnOutputOnlyStream = pa.paCanNotReadFromAnOutputOnlyStream
paCanNotWriteToAnInputOnlyStream = pa.paCanNotWriteToAnInputOnlyStream
paIncompatibleStreamHostApi = pa.paIncompatibleStreamHostApi

# PortAudio Callback Return Codes

paContinue = pa.paContinue  #: There is more audio data to come
paComplete = pa.paComplete  #: This was the last block of audio data
paAbort = pa.paAbort  #: An error ocurred, stop playback/recording

# PortAudio Callback Flags

paInputUnderflow = pa.paInputUnderflow  #: Buffer underflow in input
paInputOverflow = pa.paInputOverflow  #: Buffer overflow in input
paOutputUnderflow = pa.paOutputUnderflow  #: Buffer underflow in output
paOutputOverflow = pa.paOutputOverflow  #: Buffer overflow in output
paPrimingOutput = pa.paPrimingOutput  #: Just priming, not playing yet

# PortAudio Misc Constants

paFramesPerBufferUnspecified = pa.paFramesPerBufferUnspecified


# Utilities

def get_sample_size(format):
    """Returns the size (in bytes) for the specified sample *format*.

    :param format: A |PaSampleFormat| constant.
    :raises ValueError: on invalid specified `format`.
    :rtype: integer
    """
    return pa.get_sample_size(format)


def get_format_from_width(width, unsigned=True):
    """Returns a PortAudio format constant for the specified *width*.

    :param width: The desired sample width in bytes (1, 2, 3, or 4)
    :param unsigned: For 1 byte width, specifies signed or unsigned format.

    :raises ValueError: when invalid *width*
    :rtype: A |PaSampleFormat| constant
    """
    if width == 1:
        if unsigned:
            return paUInt8
        return paInt8
    if width == 2:
        return paInt16
    if width == 3:
        return paInt24
    if width == 4:
        return paFloat32

    raise ValueError(f"Invalid width: {width}")


# Versioning

def get_portaudio_version():
    """Returns portaudio version.

    :rtype: int
    """
    return pa.get_version()


def get_portaudio_version_text():
    """Returns PortAudio version as a text string.

    :rtype: string
    """
    return pa.get_version_text()


class PyAudio:
    """Python interface to PortAudio.

    Provides methods to:
     - initialize and terminate PortAudio
     - open and close streams
     - query and inspect the available PortAudio Host APIs
     - query and inspect the available PortAudio audio devices.

    **Stream Management**
      :py:func:`open`, :py:func:`close`

    **Host API**
      :py:func:`get_host_api_count`, :py:func:`get_default_host_api_info`,
      :py:func:`get_host_api_info_by_type`,
      :py:func:`get_host_api_info_by_index`,
      :py:func:`get_device_info_by_host_api_device_index`

    **Device API**
      :py:func:`get_device_count`, :py:func:`is_format_supported`,
      :py:func:`get_default_input_device_info`,
      :py:func:`get_default_output_device_info`,
      :py:func:`get_device_info_by_index`

    **Stream Format Conversion**
      :py:func:`get_sample_size`, :py:func:`get_format_from_width`

    **Details**
    """

    class Stream:
        """PortAudio Stream Wrapper. Use :py:func:`PyAudio.open` to instantiate.

        **Opening and Closing**
          :py:func:`__init__`, :py:func:`close`

        **Stream Info**
          :py:func:`get_input_latency`, :py:func:`get_output_latency`,
          :py:func:`get_time`, :py:func:`get_cpu_load`

        **Stream Management**
          :py:func:`start_stream`, :py:func:`stop_stream`, :py:func:`is_active`,
          :py:func:`is_stopped`

        **Input Output**
          :py:func:`write`, :py:func:`read`, :py:func:`get_read_available`,
          :py:func:`get_write_available`
        """
        def __init__(self,
                     PA_manager,
                     rate,
                     channels,
                     format,
                     input=False,
                     output=False,
                     input_device_index=None,
                     output_device_index=None,
                     frames_per_buffer=pa.paFramesPerBufferUnspecified,
                     start=True,
                     input_host_api_specific_stream_info=None,
                     output_host_api_specific_stream_info=None,
                     stream_callback=None):
            """Initialize an audio stream.

            Do not call directly. Use :py:func:`PyAudio.open`.

            A stream can either be input, output, or both.

            :param PA_manager: A reference to the managing :py:class:`PyAudio`
                instance
            :param rate: Sampling rate
            :param channels: Number of channels
            :param format: Sampling size and format. See |PaSampleFormat|.
            :param input: Specifies whether this is an input stream.
                Defaults to ``False``.
            :param output: Specifies whether this is an output stream.
                Defaults to ``False``.
            :param input_device_index: Index of Input Device to use.
                Unspecified (or ``None``) uses default device.
                Ignored if `input` is ``False``.
            :param output_device_index:
                Index of Output Device to use.
                Unspecified (or ``None``) uses the default device.
                Ignored if `output` is ``False``.
            :param frames_per_buffer: Specifies the number of frames per buffer.
            :param start: Start the stream running immediately.
                Defaults to ``True``. In general, there is no reason to set
                this to ``False``.
            :param input_host_api_specific_stream_info: Specifies a host API
                specific stream information data structure for input.

                .. only:: pamac

                   See :py:class:`PaMacCoreStreamInfo`.

            :param output_host_api_specific_stream_info: Specifies a host API
                specific stream information data structure for output.

                .. only:: pamac

                   See :py:class:`PaMacCoreStreamInfo`.

            :param stream_callback: Specifies a callback function for
                *non-blocking* (callback) operation.  Default is
                ``None``, which indicates *blocking* operation (i.e.,
                :py:func:`PyAudio.Stream.read` and
                :py:func:`PyAudio.Stream.write`).  To use non-blocking
                operation, specify a callback that conforms to the following
                signature:

                .. code-block:: python

                   callback(in_data,      # input data if input=True; else None
                            frame_count,  # number of frames
                            time_info,    # dictionary
                            status_flags) # PaCallbackFlags

                ``time_info`` is a dictionary with the following keys:
                ``input_buffer_adc_time``, ``current_time``, and
                ``output_buffer_dac_time``; see the PortAudio
                documentation for their meanings.  ``status_flags`` is one
                of |PaCallbackFlags|.

                The callback must return a tuple:

                .. code-block:: python

                    (out_data, flag)

                ``out_data`` is a byte array whose length should be the
                (``frame_count * channels * bytes-per-channel``) if
                ``output=True`` or ``None`` if ``output=False``.  ``flag``
                must be either :py:data:`paContinue`, :py:data:`paComplete` or
                :py:data:`paAbort` (one of |PaCallbackReturnCodes|).
                When ``output=True`` and ``out_data`` does not contain at
                least ``frame_count`` frames, :py:data:`paComplete` is
                assumed for ``flag``.

                **Note:** ``stream_callback`` is called in a separate
                thread (from the main thread).  Exceptions that occur in
                the ``stream_callback`` will:

                1. print a traceback on standard error to aid debugging,
                2. queue the exception to be thrown (at some point) in
                   the main thread, and
                3. return `paAbort` to PortAudio to stop the stream.

                **Note:** Do not call :py:func:`PyAudio.Stream.read` or
                :py:func:`PyAudio.Stream.write` if using non-blocking operation.

                **See:** PortAudio's callback signature for additional
                details: http://portaudio.com/docs/v19-doxydocs/portaudio_8h.html#a8a60fb2a5ec9cbade3f54a9c978e2710

            :raise ValueError: Neither input nor output are set True.
            """
            if not (input or output):
                raise ValueError("Must specify an input or output " + "stream.")

            self._parent = PA_manager
            self._is_input = input
            self._is_output = output
            self._is_running = start
            self._rate = rate
            self._channels = channels
            self._format = format
            self._frames_per_buffer = frames_per_buffer

            arguments = {
                'rate': rate,
                'channels': channels,
                'format': format,
                'input': input,
                'output': output,
                'input_device_index': input_device_index,
                'output_device_index': output_device_index,
                'frames_per_buffer': frames_per_buffer
            }

            if input_host_api_specific_stream_info:
                arguments[
                    'input_host_api_specific_stream_info'
                ] = input_host_api_specific_stream_info

            if output_host_api_specific_stream_info:
                arguments[
                    'output_host_api_specific_stream_info'
                ] = output_host_api_specific_stream_info

            if stream_callback:
                arguments['stream_callback'] = stream_callback

            # calling pa.open returns a stream object
            self._stream = pa.open(**arguments)

            self._input_latency = self._stream.inputLatency
            self._output_latency = self._stream.outputLatency

            if self._is_running:
                pa.start_stream(self._stream)

        def close(self):
            """Closes the stream."""
            pa.close(self._stream)
            self._is_running = False
            self._parent._remove_stream(self)

        # Stream Info

        def get_input_latency(self):
            """Returns the input latency.

            :rtype: float
            """
            return self._stream.inputLatency

        def get_output_latency(self):
            """Returns the output latency.

            :rtype: float
            """
            return self._stream.outputLatency

        def get_time(self):
            """Returns stream time.

            :rtype: float
            """
            return pa.get_stream_time(self._stream)

        def get_cpu_load(self):
            """Return the CPU load. Always 0.0 when using the blocking API.

            :rtype: float
            """
            return pa.get_stream_cpu_load(self._stream)

        # Stream Lifecycle

        def start_stream(self):
            """Starts the stream."""
            if self._is_running:
                return

            pa.start_stream(self._stream)
            self._is_running = True

        def stop_stream(self):
            """Stops the stream."""
            if not self._is_running:
                return

            pa.stop_stream(self._stream)
            self._is_running = False

        def is_active(self):
            """Returns whether the stream is active.

            :rtype: bool
            """
            return pa.is_stream_active(self._stream)

        def is_stopped(self):
            """Returns whether the stream is stopped.

            :rtype: bool
            """
            return pa.is_stream_stopped(self._stream)

        # Stream blocking I/O

        def write(self, frames, num_frames=None, exception_on_underflow=False):
            """Write samples to the stream for playback.

            Do not call when using non-blocking mode.

            :param frames:
               The frames of data.
            :param num_frames:
               The number of frames to write.
               Defaults to None, in which this value will be
               automatically computed.
            :param exception_on_underflow:
               Specifies whether an IOError exception should be thrown
               (or silently ignored) on buffer underflow. Defaults
               to False for improved performance, especially on
               slower platforms.

            :raises IOError: if the stream is not an output stream
               or if the write operation was unsuccessful.

            :rtype: `None`
            """
            if not self._is_output:
                raise IOError("Not output stream",
                              paCanNotWriteToAnInputOnlyStream)

            if num_frames is None:
                # Determine how many frames to read:
                width = get_sample_size(self._format)
                num_frames = int(len(frames) / (self._channels * width))

            pa.write_stream(self._stream, frames, num_frames,
                            exception_on_underflow)

        def read(self, num_frames, exception_on_overflow=True):
            """Read samples from the stream.

            Do not call when using non-blocking mode.

            :param num_frames: The number of frames to read.
            :param exception_on_overflow:
               Specifies whether an IOError exception should be thrown
               (or silently ignored) on input buffer overflow. Defaults
               to True.
            :raises IOError: if stream is not an input stream
              or if the read operation was unsuccessful.
            :rtype: bytes
            """
            if not self._is_input:
                raise IOError("Not input stream",
                              paCanNotReadFromAnOutputOnlyStream)
            return pa.read_stream(self._stream, num_frames,
                                  exception_on_overflow)

        def get_read_available(self):
            """Return the number of frames that can be read without waiting.

            :rtype: integer
            """
            return pa.get_stream_read_available(self._stream)

        def get_write_available(self):
            """Return the number of frames that can be written without waiting.

            :rtype: integer
            """
            return pa.get_stream_write_available(self._stream)

    # Initialization and Termination

    def __init__(self):
        """Initialize PortAudio."""
        pa.initialize()
        self._streams = set()

    def terminate(self):
        """Terminates PortAudio.

        :attention: Be sure to call this method for every instance of this
          object to release PortAudio resources.
        """
        for stream in self._streams.copy():
            stream.close()

        self._streams = set()
        pa.terminate()

    # Utilities

    def get_sample_size(self, format):
        """Returns the size (in bytes) for the specified sample `format`
        (a |PaSampleFormat| constant).

        :param format: A |PaSampleFormat| constant.
        :raises ValueError: Invalid specified `format`.
        :rtype: integer
        """
        return pa.get_sample_size(format)

    def get_format_from_width(self, width, unsigned=True):
        """Returns a PortAudio format constant for the specified `width`.

        :param width: The desired sample width in bytes (1, 2, 3, or 4)
        :param unsigned: For 1 byte width, specifies signed or unsigned format.

        :raises ValueError: for invalid `width`
        :rtype: A |PaSampleFormat| constant.
        """
        return get_format_from_width(width, unsigned)

    # Stream Factory

    def open(self, *args, **kwargs):
        """Opens a new stream.

        See constructor for :py:func:`PyAudio.Stream.__init__` for parameter
        details.

        :returns: A new :py:class:`PyAudio.Stream`
        """
        stream = PyAudio.Stream(self, *args, **kwargs)
        self._streams.add(stream)
        return stream

    def close(self, stream):
        """Closes a stream. Use :py:func:`PyAudio.Stream.close` instead.

        :param stream: An instance of the :py:class:`PyAudio.Stream` object.
        :raises ValueError: if stream does not exist.
        """
        if stream not in self._streams:
            raise ValueError(f"Stream {stream} not found")

        stream.close()

    def _remove_stream(self, stream):
        """Removes a stream. (Internal)

        :param stream: An instance of the :py:class:`PyAudio.Stream` object.
        """
        if stream in self._streams:
            self._streams.remove(stream)

    # Host API Inspection

    def get_host_api_count(self):
        """Returns the number of available PortAudio Host APIs.

        :rtype: integer
        """
        return pa.get_host_api_count()

    def get_default_host_api_info(self):
        """Returns a dictionary containing the default Host API parameters.

        The keys of the dictionary mirror the data fields of PortAudio's
        ``PaHostApiInfo`` structure.

        :raises IOError: if no default input device is available
        :rtype: dict
        """
        default_host_api_index = pa.get_default_host_api()
        return self.get_host_api_info_by_index(default_host_api_index)

    def get_host_api_info_by_type(self, host_api_type):
        """Returns a dictionary containing the Host API parameters for the
        host API specified by the `host_api_type`. The keys of the
        dictionary mirror the data fields of PortAudio's ``PaHostApiInfo``
        structure.

        :param host_api_type: The desired |PaHostAPI|
        :raises IOError: for invalid `host_api_type`
        :rtype: dict
        """
        index = pa.host_api_type_id_to_host_api_index(host_api_type)
        return self.get_host_api_info_by_index(index)

    def get_host_api_info_by_index(self, host_api_index):
        """Returns a dictionary containing the Host API parameters for the
        host API specified by the `host_api_index`. The keys of the
        dictionary mirror the data fields of PortAudio's ``PaHostApiInfo``
        structure.

        :param host_api_index: The host api index
        :raises IOError: for invalid `host_api_index`
        :rtype: dict
        """
        return self._make_host_api_dictionary(
            host_api_index,
            pa.get_host_api_info(host_api_index))

    def get_device_info_by_host_api_device_index(self,
                                                 host_api_index,
                                                 host_api_device_index):
        """Returns a dictionary containing the Device parameters for a
        given Host API's n'th device. The keys of the dictionary
        mirror the data fields of PortAudio's ``PaDeviceInfo`` structure.

        :param host_api_index: The Host API index number
        :param host_api_device_index: The n'th device of the host API
        :raises IOError: for invalid indices
        :rtype: dict
        """
        long_method_name = pa.host_api_device_index_to_device_index
        device_index = long_method_name(host_api_index, host_api_device_index)
        return self.get_device_info_by_index(device_index)

    def _make_host_api_dictionary(self, index, host_api_struct):
        """Creates dictionary like PortAudio's ``PaHostApiInfo`` structure.

        :rtype: dict
        """
        return {
            'index': index,
            'structVersion': host_api_struct.structVersion,
            'type': host_api_struct.type,
            'name': host_api_struct.name,
            'deviceCount': host_api_struct.deviceCount,
            'defaultInputDevice': host_api_struct.defaultInputDevice,
            'defaultOutputDevice': host_api_struct.defaultOutputDevice
        }

    # Device Inspection

    def get_device_count(self):
        """Returns the number of PortAudio Host APIs.

        :rtype: integer
        """
        return pa.get_device_count()

    def is_format_supported(self, rate,
                            input_device=None,
                            input_channels=None,
                            input_format=None,
                            output_device=None,
                            output_channels=None,
                            output_format=None):
        """Checks if specified device configuration is supported.

        Returns True if the configuration is supported; raises ValueError
        otherwise.

        :param rate:
           Specifies the desired rate (in Hz)
        :param input_device:
           The input device index. Specify ``None`` (default) for
           half-duplex output-only streams.
        :param input_channels:
           The desired number of input channels. Ignored if
           `input_device` is not specified (or ``None``).
        :param input_format:
           PortAudio sample format constant defined
           in this module
        :param output_device:
           The output device index. Specify ``None`` (default) for
           half-duplex input-only streams.
        :param output_channels:
           The desired number of output channels. Ignored if
           `input_device` is not specified (or ``None``).
        :param output_format:
           |PaSampleFormat| constant.

        :rtype: bool
        :raises ValueError: tuple containing (error string, |PaErrorCode|).
        """
        if input_device is None and output_device is None:
            raise ValueError(
                "Must specify stream format for input, output, or both",
                paInvalidDevice)

        kwargs = {}
        if input_device is not None:
            kwargs['input_device'] = input_device
            kwargs['input_channels'] = input_channels
            kwargs['input_format'] = input_format

        if output_device is not None:
            kwargs['output_device'] = output_device
            kwargs['output_channels'] = output_channels
            kwargs['output_format'] = output_format

        return pa.is_format_supported(rate, **kwargs)

    def get_default_input_device_info(self):
        """Returns the default input device parameters as a dictionary.

        The keys of the dictionary mirror the data fields of PortAudio's
        ``PaDeviceInfo`` structure.

        :raises IOError: No default input device available.
        :rtype: dict
        """
        device_index = pa.get_default_input_device()
        return self.get_device_info_by_index(device_index)

    def get_default_output_device_info(self):
        """Returns the default output device parameters as a dictionary.

        The keys of the dictionary mirror the data fields of PortAudio's
        ``PaDeviceInfo`` structure.

        :raises IOError: No default output device available.
        :rtype: dict
        """
        device_index = pa.get_default_output_device()
        return self.get_device_info_by_index(device_index)

    def get_device_info_by_index(self, device_index):
        """Returns the device parameters for device specified in `device_index`
        as a dictionary. The keys of the dictionary mirror the data fields of
        PortAudio's ``PaDeviceInfo`` structure.

        :param device_index: The device index
        :raises IOError: Invalid `device_index`.
        :rtype: dict
        """
        return self._make_device_info_dictionary(
            device_index,
            pa.get_device_info(device_index))

    def _make_device_info_dictionary(self, index, device_info):
        """Creates a dictionary like PortAudio's ``PaDeviceInfo`` structure.

        :rtype: dict
        """
        device_name = device_info.name

        # Attempt to decode device_name. If we fail to decode, return the raw
        # bytes and let the caller deal with the encoding.
        os_encoding = locale.getpreferredencoding(do_setlocale=False)
        for codec in [os_encoding, "utf-8"]:
            try:
                device_name = device_name.decode(codec)
                break
            except:
                pass

        return {'index': index,
                'structVersion': device_info.structVersion,
                'name': device_name,
                'hostApi': device_info.hostApi,
                'maxInputChannels': device_info.maxInputChannels,
                'maxOutputChannels': device_info.maxOutputChannels,
                'defaultLowInputLatency':
                device_info.defaultLowInputLatency,
                'defaultLowOutputLatency':
                device_info.defaultLowOutputLatency,
                'defaultHighInputLatency':
                device_info.defaultHighInputLatency,
                'defaultHighOutputLatency':
                device_info.defaultHighOutputLatency,
                'defaultSampleRate':
                device_info.defaultSampleRate}


# Host Specific Stream Info

if hasattr(pa, 'paMacCoreStreamInfo'):
    class PaMacCoreStreamInfo(pa.paMacCoreStreamInfo):
        """PortAudio Host API Specific Stream Info for macOS-specific settings.

        To configure macOS-specific settings, instantiate this class and pass
        it as the argument in :py:func:`PyAudio.open` to parameters
        ``input_host_api_specific_stream_info`` or
        ``output_host_api_specific_stream_info``.  (See
        :py:func:`PyAudio.Stream.__init__`.)

        :note: macOS-only.

        .. |PaMacCoreFlags| replace:: :ref:`PortAudio Mac Core Flags <PaMacCoreFlags>`
        .. _PaMacCoreFlags:

        **PortAudio Mac Core Flags**
          :py:data:`paMacCoreChangeDeviceParameters`,
          :py:data:`paMacCoreFailIfConversionRequired`,
          :py:data:`paMacCoreConversionQualityMin`,
          :py:data:`paMacCoreConversionQualityMedium`,
          :py:data:`paMacCoreConversionQualityLow`,
          :py:data:`paMacCoreConversionQualityHigh`,
          :py:data:`paMacCoreConversionQualityMax`,
          :py:data:`paMacCorePlayNice`,
          :py:data:`paMacCorePro`,
          :py:data:`paMacCoreMinimizeCPUButPlayNice`,
          :py:data:`paMacCoreMinimizeCPU`

        .. attribute:: flags

           The flags specified to the constructor.

           :type: |PaMacCoreFlags|

        .. attribute:: channel_map

           The channel_map specified to the constructor

           :type: tuple or None if unspecified
        """
        paMacCoreChangeDeviceParameters = pa.paMacCoreChangeDeviceParameters
        paMacCoreFailIfConversionRequired = pa.paMacCoreFailIfConversionRequired
        paMacCoreConversionQualityMin = pa.paMacCoreConversionQualityMin
        paMacCoreConversionQualityMedium = pa.paMacCoreConversionQualityMedium
        paMacCoreConversionQualityLow = pa.paMacCoreConversionQualityLow
        paMacCoreConversionQualityHigh = pa.paMacCoreConversionQualityHigh
        paMacCoreConversionQualityMax = pa.paMacCoreConversionQualityMax
        paMacCorePlayNice = pa.paMacCorePlayNice
        paMacCorePro = pa.paMacCorePro
        paMacCoreMinimizeCPUButPlayNice = pa.paMacCoreMinimizeCPUButPlayNice
        paMacCoreMinimizeCPU = pa.paMacCoreMinimizeCPU

        def __init__(self, flags=None, channel_map=None):
            """Initialize with macOS setting flags and channel_map.

            See PortAudio documentation for more details on these parameters.

            :param flags: |PaMacCoreFlags| OR'ed together.
            :param channel_map: An array describing the channel mapping.
                See PortAudio documentation for usage.
            """
            kwargs = {}
            if flags is not None:
                kwargs["flags"] = flags
            if channel_map is not None:
                kwargs["channel_map"] = channel_map
            super().__init__(**kwargs)

        # Deprecated:

        def get_flags(self):
            """Returns the flags set at instantiation. Deprecated.

            :rtype: integer

            .. deprecated:: 0.2.13
               Use :py:attr:`flags` property.
            """
            warnings.warn(
                "PaMacCoreStreamInfo.get_flags is deprecated. Use the flags "
                "property instead.",
                DeprecationWarning,
                stacklevel=2)
            return self.flags

        def get_channel_map(self):
            """Returns the channel map set at instantiation. Deprecated.

            :rtype: tuple or None

            .. deprecated:: 0.2.13
               Use :py:attr:`channel_map` property.
            """
            warnings.warn(
                "PaMacCoreStreamInfo.get_channel_map is deprecated. Use the "
                "channel_map property instead.",
                DeprecationWarning,
                stacklevel=2)
            return self.channel_map

        def _get_host_api_stream_object(self):
            """Returns the underyling stream info.

            .. :deprecated:: 0.2.13
               Use stream_info property.
            """
            warnings.warn(
                "PaMacCoreStreamInfo._get_host_api_stream_object is "
                "deprecated. Use this object instance instead.",
                DeprecationWarning,
                stacklevel=2)
            return self


# The top-level Stream class is reserved for future API changes. Users should
# never instantiate Stream directly. Instead, users must use PyAudio.open()
# instead, as documented.
#
# But for existing code that happens to instantiate Stream directly, this class
# issues a warning and maintains backwards-compatibility, for now. In the
# future, Stream may be repurposed.
class Stream(PyAudio.Stream):
    """Reserved. Do not instantiate."""

    def __init__(self, *args, **kwargs):
        # Users should never instantiate this class.
        warnings.warn(
            "Do not instantiate pyaudio.Stream directly. Use "
            "pyaudio.PyAudio.open() instead. pyaudio.Stream may change or be "
            "removed in the future.",
            DeprecationWarning,
            stacklevel=2)
        super().__init__(*args, **kwargs)



import doctest
import collections

# TODO - finish doc tests

# TODO - unit tests needed for get/set and Box named tuple

__version__ = "0.2.0"


# Constants for rectangle attributes:
TOP = "top"
BOTTOM = "bottom"
LEFT = "left"
RIGHT = "right"
TOPLEFT = "topleft"
TOPRIGHT = "topright"
BOTTOMLEFT = "bottomleft"
BOTTOMRIGHT = "bottomright"
MIDTOP = "midtop"
MIDRIGHT = "midright"
MIDLEFT = "midleft"
MIDBOTTOM = "midbottom"
CENTER = "center"
CENTERX = "centerx"
CENTERY = "centery"
WIDTH = "width"
HEIGHT = "height"
SIZE = "size"
BOX = "box"
AREA = "area"
PERIMETER = "perimeter"

Box = collections.namedtuple("Box", "left top width height")
Point = collections.namedtuple("Point", "x y")
Size = collections.namedtuple("Size", "width height")


class PyRectException(Exception):
    """
    This class exists for PyRect exceptions. If the PyRect module raises any
    non-PyRectException exceptions, this indicates there's a bug in PyRect.
    """

    pass


def _checkForIntOrFloat(arg):
    """Raises an exception if arg is not an int or float. Always returns None."""
    if not isinstance(arg, (int, float)):
        raise PyRectException(
            "argument must be int or float, not %s" % (arg.__class__.__name__)
        )


def _checkForInt(arg):
    """Raises an exception if arg is not an int. Always returns None."""
    if not isinstance(arg, int):
        raise PyRectException(
            "argument must be int or float, not %s" % (arg.__class__.__name__)
        )


def _checkForTwoIntOrFloatTuple(arg):
    try:
        if not isinstance(arg[0], (int, float)) or not isinstance(arg[1], (int, float)):
            raise PyRectException(
                "argument must be a two-item tuple containing int or float values"
            )
    except:
        raise PyRectException(
            "argument must be a two-item tuple containing int or float values"
        )


def _checkForFourIntOrFloatTuple(arg):
    try:
        if (
            not isinstance(arg[0], (int, float))
            or not isinstance(arg[1], (int, float))
            or not isinstance(arg[2], (int, float))
            or not isinstance(arg[3], (int, float))
        ):
            raise PyRectException(
                "argument must be a four-item tuple containing int or float values"
            )
    except:
        raise PyRectException(
            "argument must be a four-item tuple containing int or float values"
        )


def _collides(rectOrPoint1, rectOrPoint2):
    """Returns True if rectOrPoint1 and rectOrPoint2 collide with each other."""


def _getRectsAndPoints(rectsOrPoints):
    points = []
    rects = []
    for rectOrPoint in rectsOrPoints:
        try:
            _checkForTwoIntOrFloatTuple(rectOrPoint)
            points.append(rectOrPoint)
        except PyRectException:
            try:
                _checkForFourIntOrFloatTuple(rectOrPoint)
            except:
                raise PyRectException("argument is not a point or a rect tuple")
            rects.append(rectOrPoint)
    return (rects, points)


'''
def collideAnyBetween(rectsOrPoints):
    """Returns True if any of the (x, y) or (left, top, width, height) tuples
    in rectsOrPoints collides with any other point or box tuple in rectsOrPoints.

    >>> p1 = (50, 50)
    >>> p2 = (100, 100)
    >>> p3 = (50, 200)
    >>> r1 = (-50, -50, 20, 20)
    >>> r2 = (25, 25, 50, 50)
    >>> collideAnyBetween([p1, p2, p3, r1, r2]) # p1 and r2 collide
    True
    >>> collideAnyBetween([p1, p2, p3, r1])
    False
    """
    # TODO - needs to be complete

    # split up
    rects, points = _getRectsAndPoints(rectsOrPoints)

    # compare points with each other
    if len(points) > 1:
        for point in points:
            if point != points[0]:
                return False

    # TODO finish
'''


'''
def collideAllBetween(rectsOrPoints):
    """Returns True if any of the (x, y) or (left, top, width, height) tuples
    in rectsOrPoints collides with any other point or box tuple in rectsOrPoints.

    >>> p1 = (50, 50)
    >>> p2 = (100, 100)
    >>> p3 = (50, 200)
    >>> r1 = (-50, -50, 20, 20)
    >>> r2 = (25, 25, 50, 50)
    >>> collideAllBetween([p1, p2, p3, r1, r2])
    False
    >>> collideAllBetween([p1, p2, p3, r1])
    False
    >>> collideAllBetween([p1, r2]) # Everything in the list collides with each other.
    True
    """

    # Check for valid arguments
    try:
        for rectOrPoint in rectsOrPoints:
            if len(rectOrPoint) == 2:
                _checkForTwoIntOrFloatTuple(rectOrPoint)
            elif len(rectOrPoint) == 4:
                _checkForFourIntOrFloatTuple(rectOrPoint)
            else:
                raise PyRectException()
    except:
        raise PyRectException('Arguments in rectsOrPoints must be 2- or 4-integer/float tuples.')

    raise NotImplementedError # return a list of all rects or points that collide with any other in the argument
'''


class Rect(object):
    def __init__(
        self,
        left=0,
        top=0,
        width=0,
        height=0,
        enableFloat=False,
        readOnly=False,
        onChange=None,
        onRead=None,
    ):
        _checkForIntOrFloat(width)
        _checkForIntOrFloat(height)
        _checkForIntOrFloat(left)
        _checkForIntOrFloat(top)

        self._enableFloat = bool(enableFloat)
        self._readOnly = bool(readOnly)

        if onChange is not None and not callable(onChange):
            raise PyRectException(
                "onChange argument must be None or callable (function, method, etc.)"
            )
        self.onChange = onChange

        if onRead is not None and not callable(onRead):
            raise PyRectException(
                "onRead argument must be None or callable (function, method, etc.)"
            )
        self.onRead = onRead

        if enableFloat:
            self._width = float(width)
            self._height = float(height)
            self._left = float(left)
            self._top = float(top)
        else:
            self._width = int(width)
            self._height = int(height)
            self._left = int(left)
            self._top = int(top)

    # OPERATOR OVERLOADING / DUNDER METHODS
    def __repr__(self):
        """Return a string of the constructor function call to create this Rect object."""
        return "%s(left=%s, top=%s, width=%s, height=%s)" % (
            self.__class__.__name__,
            self._left,
            self._top,
            self._width,
            self._height,
        )

    def __str__(self):
        """Return a string representation of this Rect object."""
        return "(x=%s, y=%s, w=%s, h=%s)" % (
            self._left,
            self._top,
            self._width,
            self._height,
        )

    def callOnChange(self, oldLeft, oldTop, oldWidth, oldHeight):
        # Note: callOnChange() should be called *after* the attribute has been changed.
        # Note: This isn't thread safe; the attributes can change between the calling of this function and the code in the function running.
        if self.onChange is not None:
            self.onChange(
                Box(oldLeft, oldTop, oldWidth, oldHeight),
                Box(self._left, self._top, self._width, self._height),
            )

    @property
    def enableFloat(self):
        """
        A Boolean attribute that determines if this rectangle uses floating point
        numbers for its position and size. False, by default.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.enableFloat
        False
        >>> r.enableFloat = True
        >>> r.top = 3.14
        >>> r
        Rect(left=0.0, top=3.14, width=10.0, height=20.0)
        """
        return self._enableFloat

    @enableFloat.setter
    def enableFloat(self, value):
        if not isinstance(value, bool):
            raise PyRectException("enableFloat must be set to a bool value")
        self._enableFloat = value

        if self._enableFloat:
            self._left = float(self._left)
            self._top = float(self._top)
            self._width = float(self._width)
            self._height = float(self._height)
        else:
            self._left = int(self._left)
            self._top = int(self._top)
            self._width = int(self._width)
            self._height = int(self._height)

    # LEFT SIDE PROPERTY
    @property
    def left(self):
        """
        The x coordinate for the left edge of the rectangle. `x` is an alias for `left`.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.left
        0
        >>> r.left = 50
        >>> r
        Rect(left=50, top=0, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(LEFT)
        return self._left

    @left.setter
    def left(self, newLeft):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(newLeft)
        if (
            newLeft != self._left
        ):  # Only run this code if the size/position has changed.
            originalLeft = self._left
            if self._enableFloat:
                self._left = newLeft
            else:
                self._left = int(newLeft)
            self.callOnChange(originalLeft, self._top, self._width, self._height)

    x = left  # x is an alias for left

    # TOP SIDE PROPERTY
    @property
    def top(self):
        """
        The y coordinate for the top edge of the rectangle. `y` is an alias for `top`.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.top
        0
        >>> r.top = 50
        >>> r
        Rect(left=0, top=50, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(TOP)
        return self._top

    @top.setter
    def top(self, newTop):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(newTop)
        if newTop != self._top:  # Only run this code if the size/position has changed.
            originalTop = self._top
            if self._enableFloat:
                self._top = newTop
            else:
                self._top = int(newTop)
            self.callOnChange(self._left, originalTop, self._width, self._height)

    y = top  # y is an alias for top

    # RIGHT SIDE PROPERTY
    @property
    def right(self):
        """
        The x coordinate for the right edge of the rectangle.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.right
        10
        >>> r.right = 50
        >>> r
        Rect(left=40, top=0, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(RIGHT)
        return self._left + self._width

    @right.setter
    def right(self, newRight):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(newRight)
        if (
            newRight != self._left + self._width
        ):  # Only run this code if the size/position has changed.
            originalLeft = self._left
            if self._enableFloat:
                self._left = newRight - self._width
            else:
                self._left = int(newRight) - self._width
            self.callOnChange(originalLeft, self._top, self._width, self._height)

    # BOTTOM SIDE PROPERTY
    @property
    def bottom(self):
        """The y coordinate for the bottom edge of the rectangle.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.bottom
        20
        >>> r.bottom = 30
        >>> r
        Rect(left=0, top=10, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(BOTTOM)
        return self._top + self._height

    @bottom.setter
    def bottom(self, newBottom):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(newBottom)
        if (
            newBottom != self._top + self._height
        ):  # Only run this code if the size/position has changed.
            originalTop = self._top
            if self._enableFloat:
                self._top = newBottom - self._height
            else:
                self._top = int(newBottom) - self._height
            self.callOnChange(self._left, originalTop, self._width, self._height)

    # TOP LEFT CORNER PROPERTY
    @property
    def topleft(self):
        """
        The x and y coordinates for the top right corner of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.topleft
        (0, 0)
        >>> r.topleft = (30, 30)
        >>> r
        Rect(left=30, top=30, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(TOPLEFT)
        return Point(x=self._left, y=self._top)

    @topleft.setter
    def topleft(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newLeft, newTop = value
        if (newLeft != self._left) or (
            newTop != self._top
        ):  # Only run this code if the size/position has changed.
            originalLeft = self._left
            originalTop = self._top
            if self._enableFloat:
                self._left = newLeft
                self._top = newTop
            else:
                self._left = int(newLeft)
                self._top = int(newTop)
            self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # BOTTOM LEFT CORNER PROPERTY
    @property
    def bottomleft(self):
        """
        The x and y coordinates for the bottom right corner of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.bottomleft
        (0, 20)
        >>> r.bottomleft = (30, 30)
        >>> r
        Rect(left=30, top=10, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(BOTTOMLEFT)
        return Point(x=self._left, y=self._top + self._height)

    @bottomleft.setter
    def bottomleft(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newLeft, newBottom = value
        if (newLeft != self._left) or (
            newBottom != self._top + self._height
        ):  # Only run this code if the size/position has changed.
            originalLeft = self._left
            originalTop = self._top
            if self._enableFloat:
                self._left = newLeft
                self._top = newBottom - self._height
            else:
                self._left = int(newLeft)
                self._top = int(newBottom) - self._height
            self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # TOP RIGHT CORNER PROPERTY
    @property
    def topright(self):
        """
        The x and y coordinates for the top right corner of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.topright
        (10, 0)
        >>> r.topright = (30, 30)
        >>> r
        Rect(left=20, top=30, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(TOPRIGHT)
        return Point(x=self._left + self._width, y=self._top)

    @topright.setter
    def topright(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newRight, newTop = value
        if (newRight != self._left + self._width) or (
            newTop != self._top
        ):  # Only run this code if the size/position has changed.
            originalLeft = self._left
            originalTop = self._top
            if self._enableFloat:
                self._left = newRight - self._width
                self._top = newTop
            else:
                self._left = int(newRight) - self._width
                self._top = int(newTop)
            self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # BOTTOM RIGHT CORNER PROPERTY
    @property
    def bottomright(self):
        """
        The x and y coordinates for the bottom right corner of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.bottomright
        (10, 20)
        >>> r.bottomright = (30, 30)
        >>> r
        Rect(left=20, top=10, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(BOTTOMRIGHT)
        return Point(x=self._left + self._width, y=self._top + self._height)

    @bottomright.setter
    def bottomright(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newRight, newBottom = value
        if (newBottom != self._top + self._height) or (
            newRight != self._left + self._width
        ):  # Only run this code if the size/position has changed.
            originalLeft = self._left
            originalTop = self._top
            if self._enableFloat:
                self._left = newRight - self._width
                self._top = newBottom - self._height
            else:
                self._left = int(newRight) - self._width
                self._top = int(newBottom) - self._height
            self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # MIDDLE OF TOP SIDE PROPERTY
    @property
    def midtop(self):
        """
        The x and y coordinates for the midpoint of the top edge of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.midtop
        (5, 0)
        >>> r.midtop = (40, 50)
        >>> r
        Rect(left=35, top=50, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(MIDTOP)
        if self._enableFloat:
            return Point(x=self._left + (self._width / 2.0), y=self._top)
        else:
            return Point(x=self._left + (self._width // 2), y=self._top)

    @midtop.setter
    def midtop(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newMidTop, newTop = value
        originalLeft = self._left
        originalTop = self._top
        if self._enableFloat:
            if (newMidTop != self._left + self._width / 2.0) or (
                newTop != self._top
            ):  # Only run this code if the size/position has changed.
                self._left = newMidTop - (self._width / 2.0)
                self._top = newTop
                self.callOnChange(originalLeft, originalTop, self._width, self._height)
        else:
            if (newMidTop != self._left + self._width // 2) or (
                newTop != self._top
            ):  # Only run this code if the size/position has changed.
                self._left = int(newMidTop) - (self._width // 2)
                self._top = int(newTop)
                self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # MIDDLE OF BOTTOM SIDE PROPERTY
    @property
    def midbottom(self):
        """
        The x and y coordinates for the midpoint of the bottom edge of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.midbottom
        (5, 20)
        >>> r.midbottom = (40, 50)
        >>> r
        Rect(left=35, top=30, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(MIDBOTTOM)
        if self._enableFloat:
            return Point(x=self._left + (self._width / 2.0), y=self._top + self._height)
        else:
            return Point(x=self._left + (self._width // 2), y=self._top + self._height)

    @midbottom.setter
    def midbottom(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newMidBottom, newBottom = value
        originalLeft = self._left
        originalTop = self._top
        if self._enableFloat:
            if (newMidBottom != self._left + self._width / 2.0) or (
                newBottom != self._top + self._height
            ):  # Only run this code if the size/position has changed.
                self._left = newMidBottom - (self._width / 2.0)
                self._top = newBottom - self._height
                self.callOnChange(originalLeft, originalTop, self._width, self._height)
        else:
            if (newMidBottom != self._left + self._width // 2) or (
                newBottom != self._top + self._height
            ):  # Only run this code if the size/position has changed.
                self._left = int(newMidBottom) - (self._width // 2)
                self._top = int(newBottom) - self._height
                self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # MIDDLE OF LEFT SIDE PROPERTY
    @property
    def midleft(self):
        """
        The x and y coordinates for the midpoint of the left edge of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.midleft
        (0, 10)
        >>> r.midleft = (40, 50)
        >>> r
        Rect(left=40, top=40, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(MIDLEFT)
        if self._enableFloat:
            return Point(x=self._left, y=self._top + (self._height / 2.0))
        else:
            return Point(x=self._left, y=self._top + (self._height // 2))

    @midleft.setter
    def midleft(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newLeft, newMidLeft = value
        originalLeft = self._left
        originalTop = self._top
        if self._enableFloat:
            if (newLeft != self._left) or (
                newMidLeft != self._top + (self._height / 2.0)
            ):  # Only run this code if the size/position has changed.
                self._left = newLeft
                self._top = newMidLeft - (self._height / 2.0)
                self.callOnChange(originalLeft, originalTop, self._width, self._height)
        else:
            if (newLeft != self._left) or (
                newMidLeft != self._top + (self._height // 2)
            ):  # Only run this code if the size/position has changed.
                self._left = int(newLeft)
                self._top = int(newMidLeft) - (self._height // 2)
                self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # MIDDLE OF RIGHT SIDE PROPERTY
    @property
    def midright(self):
        """
        The x and y coordinates for the midpoint of the right edge of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.midright
        (10, 10)
        >>> r.midright = (40, 50)
        >>> r
        Rect(left=30, top=40, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(MIDRIGHT)
        if self._enableFloat:
            return Point(x=self._left + self._width, y=self._top + (self._height / 2.0))
        else:
            return Point(x=self._left + self._width, y=self._top + (self._height // 2))

    @midright.setter
    def midright(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newRight, newMidRight = value
        originalLeft = self._left
        originalTop = self._top
        if self._enableFloat:
            if (newRight != self._left + self._width) or (
                newMidRight != self._top + self._height / 2.0
            ):  # Only run this code if the size/position has changed.
                self._left = newRight - self._width
                self._top = newMidRight - (self._height / 2.0)
                self.callOnChange(originalLeft, originalTop, self._width, self._height)
        else:
            if (newRight != self._left + self._width) or (
                newMidRight != self._top + self._height // 2
            ):  # Only run this code if the size/position has changed.
                self._left = int(newRight) - self._width
                self._top = int(newMidRight) - (self._height // 2)
                self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # CENTER POINT PROPERTY
    @property
    def center(self):
        """
        The x and y coordinates for the center of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.center
        (5, 10)
        >>> r.center = (40, 50)
        >>> r
        Rect(left=35, top=40, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(CENTER)
        if self._enableFloat:
            return Point(
                x=self._left + (self._width / 2.0), y=self._top + (self._height / 2.0)
            )
        else:
            return Point(
                x=self._left + (self._width // 2), y=self._top + (self._height // 2)
            )

    @center.setter
    def center(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newCenterx, newCentery = value
        originalLeft = self._left
        originalTop = self._top
        if self._enableFloat:
            if (newCenterx != self._left + self._width / 2.0) or (
                newCentery != self._top + self._height / 2.0
            ):  # Only run this code if the size/position has changed.
                self._left = newCenterx - (self._width / 2.0)
                self._top = newCentery - (self._height / 2.0)
                self.callOnChange(originalLeft, originalTop, self._width, self._height)
        else:
            if (newCenterx != self._left + self._width // 2) or (
                newCentery != self._top + self._height // 2
            ):  # Only run this code if the size/position has changed.
                self._left = int(newCenterx) - (self._width // 2)
                self._top = int(newCentery) - (self._height // 2)
                self.callOnChange(originalLeft, originalTop, self._width, self._height)

    # X COORDINATE OF CENTER POINT PROPERTY
    @property
    def centerx(self):
        """
        The x coordinate for the center of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.centerx
        5
        >>> r.centerx = 50
        >>> r
        Rect(left=45, top=0, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(CENTERX)
        if self._enableFloat:
            return self._left + (self._width / 2.0)
        else:
            return self._left + (self._width // 2)

    @centerx.setter
    def centerx(self, newCenterx):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(newCenterx)
        originalLeft = self._left
        if self._enableFloat:
            if (
                newCenterx != self._left + self._width / 2.0
            ):  # Only run this code if the size/position has changed.
                self._left = newCenterx - (self._width / 2.0)
                self.callOnChange(originalLeft, self._top, self._width, self._height)
        else:
            if (
                newCenterx != self._left + self._width // 2
            ):  # Only run this code if the size/position has changed.
                self._left = int(newCenterx) - (self._width // 2)
                self.callOnChange(originalLeft, self._top, self._width, self._height)

    # Y COORDINATE OF CENTER POINT PROPERTY
    @property
    def centery(self):
        """
        The y coordinate for the center of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.centery
        10
        >>> r.centery = 50
        >>> r
        Rect(left=0, top=40, width=10, height=20)
        """
        if self.onRead is not None:
            self.onRead(CENTERY)
        if self._enableFloat:
            return self._top + (self._height / 2.0)
        else:
            return self._top + (self._height // 2)

    @centery.setter
    def centery(self, newCentery):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(newCentery)
        originalTop = self._top
        if self._enableFloat:
            if (
                newCentery != self._top + self._height / 2.0
            ):  # Only run this code if the size/position has changed.
                self._top = newCentery - (self._height / 2.0)
                self.callOnChange(self._left, originalTop, self._width, self._height)
        else:
            if (
                newCentery != self._top + self._height // 2
            ):  # Only run this code if the size/position has changed.
                self._top = int(newCentery) - (self._height // 2)
                self.callOnChange(self._left, originalTop, self._width, self._height)

    # SIZE PROPERTY (i.e. (width, height))
    @property
    def size(self):
        """
        The width and height of the rectangle, as a tuple.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.size
        (10, 20)
        >>> r.size = (40, 50)
        >>> r
        Rect(left=0, top=0, width=40, height=50)
        """
        if self.onRead is not None:
            self.onRead(SIZE)
        return Size(width=self._width, height=self._height)

    @size.setter
    def size(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForTwoIntOrFloatTuple(value)
        newWidth, newHeight = value
        if newWidth != self._width or newHeight != self._height:
            originalWidth = self._width
            originalHeight = self._height
            if self._enableFloat:
                self._width = newWidth
                self._height = newHeight
            else:
                self._width = int(newWidth)
                self._height = int(newHeight)
            self.callOnChange(self._left, self._top, originalWidth, originalHeight)

    # WIDTH PROPERTY
    @property
    def width(self):
        """
        The width of the rectangle. `w` is an alias for `width`.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.width
        10
        >>> r.width = 50
        >>> r
        Rect(left=0, top=0, width=50, height=20)
        """
        if self.onRead is not None:
            self.onRead(WIDTH)
        return self._width

    @width.setter
    def width(self, newWidth):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(newWidth)
        if (
            newWidth != self._width
        ):  # Only run this code if the size/position has changed.
            originalWidth = self._width
            if self._enableFloat:
                self._width = newWidth
            else:
                self._width = int(newWidth)
            self.callOnChange(self._left, self._top, originalWidth, self._height)

    w = width

    # HEIGHT PROPERTY
    @property
    def height(self):
        """
        The height of the rectangle. `h` is an alias for `height`

        >>> r = Rect(0, 0, 10, 20)
        >>> r.height
        20
        >>> r.height = 50
        >>> r
        Rect(left=0, top=0, width=10, height=50)
        """
        if self.onRead is not None:
            self.onRead(HEIGHT)
        return self._height

    @height.setter
    def height(self, newHeight):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(newHeight)
        if (
            newHeight != self._height
        ):  # Only run this code if the size/position has changed.
            originalHeight = self._height
            if self._enableFloat:
                self._height = newHeight
            else:
                self._height = int(newHeight)
            self.callOnChange(self._left, self._top, self._width, originalHeight)

    h = height

    # AREA PROPERTY
    @property
    def area(self):
        """The area of the `Rect`, which is simply the width times the height.
        This is a read-only attribute.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.area
        200
        """
        if self.onRead is not None:
            self.onRead(AREA)
        return self._width * self._height


    # PERIMETER PROPERTY
    @property
    def perimeter(self):
        """The perimeter of the `Rect`, which is simply the (width + height) * 2.
        This is a read-only attribute.

        >>> r = Rect(0, 0, 10, 20)
        >>> r.area
        200
        """
        if self.onRead is not None:
            self.onRead(AREA)
        return (self._width + self._height) * 2


    # BOX PROPERTY
    @property
    def box(self):
        """A tuple of four integers: (left, top, width, height).

        >>> r = Rect(0, 0, 10, 20)
        >>> r.box
        (0, 0, 10, 20)
        >>> r.box = (5, 15, 100, 200)
        >>> r.box
        (5, 15, 100, 200)"""
        if self.onRead is not None:
            self.onRead(BOX)
        return Box(
            left=self._left, top=self._top, width=self._width, height=self._height
        )

    @box.setter
    def box(self, value):
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForFourIntOrFloatTuple(value)
        newLeft, newTop, newWidth, newHeight = value
        if (
            (newLeft != self._left)
            or (newTop != self._top)
            or (newWidth != self._width)
            or (newHeight != self._height)
        ):
            originalLeft = self._left
            originalTop = self._top
            originalWidth = self._width
            originalHeight = self._height
            if self._enableFloat:
                self._left = float(newLeft)
                self._top = float(newTop)
                self._width = float(newWidth)
                self._height = float(newHeight)
            else:
                self._left = int(newLeft)
                self._top = int(newTop)
                self._width = int(newWidth)
                self._height = int(newHeight)
            self.callOnChange(originalLeft, originalTop, originalWidth, originalHeight)

    def get(self, rectAttrName):
        # Access via the properties so that it triggers onRead().
        if rectAttrName == TOP:
            return self.top
        elif rectAttrName == BOTTOM:
            return self.bottom
        elif rectAttrName == LEFT:
            return self.left
        elif rectAttrName == RIGHT:
            return self.right
        elif rectAttrName == TOPLEFT:
            return self.topleft
        elif rectAttrName == TOPRIGHT:
            return self.topright
        elif rectAttrName == BOTTOMLEFT:
            return self.bottomleft
        elif rectAttrName == BOTTOMRIGHT:
            return self.bottomright
        elif rectAttrName == MIDTOP:
            return self.midtop
        elif rectAttrName == MIDBOTTOM:
            return self.midbottom
        elif rectAttrName == MIDLEFT:
            return self.midleft
        elif rectAttrName == MIDRIGHT:
            return self.midright
        elif rectAttrName == CENTER:
            return self.center
        elif rectAttrName == CENTERX:
            return self.centerx
        elif rectAttrName == CENTERY:
            return self.centery
        elif rectAttrName == WIDTH:
            return self.width
        elif rectAttrName == HEIGHT:
            return self.height
        elif rectAttrName == SIZE:
            return self.size
        elif rectAttrName == AREA:
            return self.area
        elif rectAttrName == BOX:
            return self.box
        else:
            raise PyRectException("'%s' is not a valid attribute name" % (rectAttrName))

    def set(self, rectAttrName, value):
        # Set via the properties so that it triggers onChange().
        if rectAttrName == TOP:
            self.top = value
        elif rectAttrName == BOTTOM:
            self.bottom = value
        elif rectAttrName == LEFT:
            self.left = value
        elif rectAttrName == RIGHT:
            self.right = value
        elif rectAttrName == TOPLEFT:
            self.topleft = value
        elif rectAttrName == TOPRIGHT:
            self.topright = value
        elif rectAttrName == BOTTOMLEFT:
            self.bottomleft = value
        elif rectAttrName == BOTTOMRIGHT:
            self.bottomright = value
        elif rectAttrName == MIDTOP:
            self.midtop = value
        elif rectAttrName == MIDBOTTOM:
            self.midbottom = value
        elif rectAttrName == MIDLEFT:
            self.midleft = value
        elif rectAttrName == MIDRIGHT:
            self.midright = value
        elif rectAttrName == CENTER:
            self.center = value
        elif rectAttrName == CENTERX:
            self.centerx = value
        elif rectAttrName == CENTERY:
            self.centery = value
        elif rectAttrName == WIDTH:
            self.width = value
        elif rectAttrName == HEIGHT:
            self.height = value
        elif rectAttrName == SIZE:
            self.size = value
        elif rectAttrName == AREA:
            raise PyRectException("area is a read-only attribute")
        elif rectAttrName == BOX:
            self.box = value
        else:
            raise PyRectException("'%s' is not a valid attribute name" % (rectAttrName))

    def move(self, xOffset, yOffset):
        """Moves this Rect object by the given offsets. The xOffset and yOffset
        arguments can be any integer value, positive or negative.
        >>> r = Rect(0, 0, 100, 100)
        >>> r.move(10, 20)
        >>> r
        Rect(left=10, top=20, width=100, height=100)
        """
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        _checkForIntOrFloat(xOffset)
        _checkForIntOrFloat(yOffset)
        if self._enableFloat:
            self._left += xOffset
            self._top += yOffset
        else:
            self._left += int(xOffset)
            self._top += int(yOffset)

    def copy(self):
        """Return a copied `Rect` object with the same position and size as this
        `Rect` object.

        >>> r1 = Rect(0, 0, 100, 150)
        >>> r2 = r1.copy()
        >>> r1 == r2
        True
        >>> r2
        Rect(left=0, top=0, width=100, height=150)
        """
        return Rect(
            self._left,
            self._top,
            self._width,
            self._height,
            self._enableFloat,
            self._readOnly,
        )

    def inflate(self, widthChange=0, heightChange=0):
        """Increases the size of this Rect object by the given offsets. The
        rectangle's center doesn't move. Negative values will shrink the
        rectangle.

        >>> r = Rect(0, 0, 100, 150)
        >>> r.inflate(20, 40)
        >>> r
        Rect(left=-10, top=-20, width=120, height=190)
        """
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        originalCenter = self.center
        self.width += widthChange
        self.height += heightChange
        self.center = originalCenter

    def clamp(self, otherRect):
        """Centers this Rect object at the center of otherRect.

        >>> r1 =Rect(0, 0, 100, 100)
        >>> r2 = Rect(-20, -90, 50, 50)
        >>> r2.clamp(r1)
        >>> r2
        Rect(left=25, top=25, width=50, height=50)
        >>> r1.center == r2.center
        True
        """
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        self.center = otherRect.center

    '''
    def intersection(self, otherRect):
        """Returns a new Rect object of the overlapping area between this
        Rect object and otherRect.

        `clip()` is an alias for `intersection()`.
        """
        pass

    clip = intersection
    '''

    def union(self, otherRect):
        """Adjusts the width and height to also cover the area of `otherRect`.

        >>> r1 = Rect(0, 0, 100, 100)
        >>> r2 = Rect(-10, -10, 100, 100)
        >>> r1.union(r2)
        >>> r1
        Rect(left=-10, top=-10, width=110, height=110)
        """

        # TODO - Change otherRect so that it could be a point as well.

        unionLeft = min(self._left, otherRect._left)
        unionTop = min(self._top, otherRect._top)
        unionRight = max(self.right, otherRect.right)
        unionBottom = max(self.bottom, otherRect.bottom)

        self._left = unionLeft
        self._top = unionTop
        self._width = unionRight - unionLeft
        self._height = unionBottom - unionTop

    def unionAll(self, otherRects):
        """Adjusts the width and height to also cover all the `Rect` objects in
        the `otherRects` sequence.

        >>> r = Rect(0, 0, 100, 100)
        >>> r1 = Rect(0, 0, 150, 100)
        >>> r2 = Rect(-10, -10, 100, 100)
        >>> r.unionAll([r1, r2])
        >>> r
        Rect(left=-10, top=-10, width=160, height=110)
        """

        # TODO - Change otherRect so that it could be a point as well.

        otherRects = list(otherRects)
        otherRects.append(self)

        unionLeft = min([r._left for r in otherRects])
        unionTop = min([r._top for r in otherRects])
        unionRight = max([r.right for r in otherRects])
        unionBottom = max([r.bottom for r in otherRects])

        self._left = unionLeft
        self._top = unionTop
        self._width = unionRight - unionLeft
        self._height = unionBottom - unionTop

    """
    def fit(self, other):
        pass # TODO - needs to be complete
    """

    def normalize(self):
        """Rect objects with a negative width or height cover a region where the
        right/bottom edge is to the left/above of the left/top edge, respectively.
        The `normalize()` method sets the `width` and `height` to positive if they
        were negative.

        The Rect stays in the same place, though with the `top` and `left`
        attributes representing the true top and left side.

        >>> r = Rect(0, 0, -10, -20)
        >>> r.normalize()
        >>> r
        Rect(left=-10, top=-20, width=10, height=20)
        """
        if self._readOnly:
            raise PyRectException("Rect object is read-only")

        if self._width < 0:
            self._width = -self._width
            self._left -= self._width
        if self._height < 0:
            self._height = -self._height
            self._top -= self._height
        # Note: No need to intify here, since the four attributes should already be ints and no multiplication was done.

    def __contains__(
        self, value
    ):  # for either points or other Rect objects. For Rects, the *entire* Rect must be in this Rect.
        if isinstance(value, Rect):
            return (
                value.topleft in self
                and value.topright in self
                and value.bottomleft in self
                and value.bottomright in self
            )

        # Check if value is an (x, y) sequence or a (left, top, width, height) sequence.
        try:
            len(value)
        except:
            raise PyRectException(
                "in <Rect> requires an (x, y) tuple, a (left, top, width, height) tuple, or a Rect object as left operand, not %s"
                % (value.__class__.__name__)
            )

        if len(value) == 2:
            # Assume that value is an (x, y) sequence.
            _checkForTwoIntOrFloatTuple(value)
            x, y = value
            return (
                self._left < x < self._left + self._width
                and self._top < y < self._top + self._height
            )

        elif len(value) == 4:
            # Assume that value is an (x, y) sequence.
            _checkForFourIntOrFloatTuple(value)
            left, top, width, height = value
            return (
                (left, top) in self
                and (left + width, top) in self
                and (left, top + height) in self
                and (left + width, top + height) in self
            )
        else:
            raise PyRectException(
                "in <Rect> requires an (x, y) tuple, a (left, top, width, height) tuple, or a Rect object as left operand, not %s"
                % (value.__class__.__name__)
            )

    def collide(self, value):
        """Returns `True` if value collides with this `Rect` object, where value can
        be an (x, y) tuple, a (left, top, width, height) box tuple, or another `Rect`
        object. If value represents a rectangular area, any part of that area
        can collide with this `Rect` object to make `collide()` return `True`.
        Otherwise, returns `False`."""

        # Note: This code is similar to __contains__(), with some minor changes
        # because __contains__() requires the rectangular are to be COMPELTELY
        # within the Rect object.
        if isinstance(value, Rect):
            return (
                value.topleft in self
                or value.topright in self
                or value.bottomleft in self
                or value.bottomright in self
            )

        # Check if value is an (x, y) sequence or a (left, top, width, height) sequence.
        try:
            len(value)
        except:
            raise PyRectException(
                "in <Rect> requires an (x, y) tuple, a (left, top, width, height) tuple, or a Rect object as left operand, not %s"
                % (value.__class__.__name__)
            )

        if len(value) == 2:
            # Assume that value is an (x, y) sequence.
            _checkForTwoIntOrFloatTuple(value)
            x, y = value
            return (
                self._left < x < self._left + self._width
                and self._top < y < self._top + self._height
            )

        elif len(value) == 4:
            # Assume that value is an (x, y) sequence.
            left, top, width, height = value
            return (
                (left, top) in self
                or (left + width, top) in self
                or (left, top + height) in self
                or (left + width, top + height) in self
            )
        else:
            raise PyRectException(
                "in <Rect> requires an (x, y) tuple, a (left, top, width, height) tuple, or a Rect object as left operand, not %s"
                % (value.__class__.__name__)
            )

    '''
    def collideAny(self, rectsOrPoints):
        """Returns True if any of the (x, y) or (left, top, width, height)
        tuples in rectsOrPoints is inside this Rect object.

        >> r = Rect(0, 0, 100, 100)
        >> p1 = (150, 80)
        >> p2 = (100, 100) # This point collides.
        >> r.collideAny([p1, p2])
        True
        >> r1 = Rect(50, 50, 10, 20) # This Rect collides.
        >> r.collideAny([r1])
        True
        >> r.collideAny([p1, p2, r1])
        True
        """
        # TODO - needs to be complete
        pass # returns True or False
        raise NotImplementedError
'''

    '''
    def collideAll(self, rectsOrPoints):
        """Returns True if all of the (x, y) or (left, top, width, height)
        tuples in rectsOrPoints is inside this Rect object.
        """

        pass # return a list of all rects or points that collide with any other in the argument
        raise NotImplementedError
'''

    # TODO - Add overloaded operators for + - * / and others once we can determine actual use cases for them.

    """NOTE: All of the comparison magic methods compare the box tuple of Rect
    objects. This is the behavior of the pygame Rect objects. Originally,
    I thought about having the <, <=, >, and >= operators compare the area
    of Rect objects. But at the same time, I wanted to have == and != compare
    not just area, but all four left, top, width, and height attributes.
    But that's weird to have different comparison operators comparing different
    features of a rectangular area. So I just defaulted to what Pygame does
    and compares the box tuple. This means that the == and != operators are
    the only really useful comparison operators, so I decided to ditch the
    other operators altogether and just have Rect only support == and !=.
    """

    def __eq__(self, other):
        if isinstance(other, Rect):
            return other.box == self.box
        else:
            raise PyRectException(
                "Rect objects can only be compared with other Rect objects"
            )

    def __ne__(self, other):
        if isinstance(other, Rect):
            return other.box != self.box
        else:
            raise PyRectException(
                "Rect objects can only be compared with other Rect objects"
            )


if __name__ == "__main__":
    print(doctest.testmod())



from socket import socket, AF_INET, SOCK_DGRAM

DEFAULT_SERVER = ('127.0.0.1', 11113) # The default Python3 server.


class Client():
    '''
    Simple client written in Python2. Polls a server and waits for its response.
    '''
    def __init__(self):
        self.socket = socket(AF_INET, SOCK_DGRAM)
        self.socket.bind(('127.0.0.1', 0)) # OS chooses port
        self.socket_name = self.socket.getsockname()
    
    def poll_server(self, data, server=DEFAULT_SERVER):
        self.socket.sendto(bytes(data), server)
        response, addr = self.socket.recvfrom(1024)
        self.socket.close()
        return response, addr


if __name__ == '__main__':
    test_client = Client()
    print test_client.poll_server("Hello world")


# -*- coding: utf-8 -*-
import ctypes
import time
from ctypes import c_short, c_char, c_uint8, c_int32, c_int, c_uint, c_uint32, c_long, byref, Structure, CFUNCTYPE, POINTER
from ctypes.wintypes import DWORD, BOOL, HHOOK, MSG, LPWSTR, WCHAR, WPARAM, LPARAM
LPMSG = POINTER(MSG)

import atexit

from ._mouse_event import ButtonEvent, WheelEvent, MoveEvent, LEFT, RIGHT, MIDDLE, X, X2, UP, DOWN, DOUBLE, WHEEL, HORIZONTAL, VERTICAL

#https://github.com/boppreh/mouse/issues/1
#user32 = ctypes.windll.user32
user32 = ctypes.WinDLL('user32', use_last_error = True)

class MSLLHOOKSTRUCT(Structure):
    _fields_ = [("x", c_long),
                ("y", c_long),
                ('data', c_int32),
                ('reserved', c_int32),
                ("flags", DWORD),
                ("time", c_int),
                ]

LowLevelMouseProc = CFUNCTYPE(c_int, WPARAM, LPARAM, POINTER(MSLLHOOKSTRUCT))

SetWindowsHookEx = user32.SetWindowsHookExA
#SetWindowsHookEx.argtypes = [c_int, LowLevelMouseProc, c_int, c_int]
SetWindowsHookEx.restype = HHOOK

CallNextHookEx = user32.CallNextHookEx
#CallNextHookEx.argtypes = [c_int , c_int, c_int, POINTER(MSLLHOOKSTRUCT)]
CallNextHookEx.restype = c_int

UnhookWindowsHookEx = user32.UnhookWindowsHookEx
UnhookWindowsHookEx.argtypes = [HHOOK]
UnhookWindowsHookEx.restype = BOOL

GetMessage = user32.GetMessageW
GetMessage.argtypes = [LPMSG, c_int, c_int, c_int]
GetMessage.restype = BOOL

TranslateMessage = user32.TranslateMessage
TranslateMessage.argtypes = [LPMSG]
TranslateMessage.restype = BOOL

DispatchMessage = user32.DispatchMessageA
DispatchMessage.argtypes = [LPMSG]

# Beware, as of 2016-01-30 the official docs have a very incomplete list.
# This one was compiled from experience and may be incomplete.
WM_MOUSEMOVE = 0x200
WM_LBUTTONDOWN = 0x201
WM_LBUTTONUP = 0x202
WM_LBUTTONDBLCLK = 0x203
WM_RBUTTONDOWN = 0x204
WM_RBUTTONUP = 0x205
WM_RBUTTONDBLCLK = 0x206
WM_MBUTTONDOWN = 0x207
WM_MBUTTONUP = 0x208
WM_MBUTTONDBLCLK = 0x209
WM_MOUSEWHEEL = 0x20A
WM_XBUTTONDOWN = 0x20B
WM_XBUTTONUP = 0x20C
WM_XBUTTONDBLCLK = 0x20D
WM_NCXBUTTONDOWN = 0x00AB
WM_NCXBUTTONUP = 0x00AC
WM_NCXBUTTONDBLCLK = 0x00AD
WM_MOUSEHWHEEL = 0x20E
WM_LBUTTONDOWN = 0x0201
WM_LBUTTONUP = 0x0202
WM_MOUSEMOVE = 0x0200
WM_MOUSEWHEEL = 0x020A
WM_MOUSEHWHEEL = 0x020E
WM_RBUTTONDOWN = 0x0204
WM_RBUTTONUP = 0x0205

buttons_by_wm_code = {
    WM_LBUTTONDOWN: (DOWN, LEFT),
    WM_LBUTTONUP: (UP, LEFT),
    WM_LBUTTONDBLCLK: (DOUBLE, LEFT),

    WM_RBUTTONDOWN: (DOWN, RIGHT),
    WM_RBUTTONUP: (UP, RIGHT),
    WM_RBUTTONDBLCLK: (DOUBLE, RIGHT),

    WM_MBUTTONDOWN: (DOWN, MIDDLE),
    WM_MBUTTONUP: (UP, MIDDLE),
    WM_MBUTTONDBLCLK: (DOUBLE, MIDDLE),

    WM_XBUTTONDOWN: (DOWN, X),
    WM_XBUTTONUP: (UP, X),
    WM_XBUTTONDBLCLK: (DOUBLE, X),
}

MOUSEEVENTF_ABSOLUTE = 0x8000
MOUSEEVENTF_MOVE = 0x1
MOUSEEVENTF_WHEEL = 0x800
MOUSEEVENTF_HWHEEL = 0x1000
MOUSEEVENTF_LEFTDOWN = 0x2
MOUSEEVENTF_LEFTUP = 0x4
MOUSEEVENTF_RIGHTDOWN = 0x8
MOUSEEVENTF_RIGHTUP = 0x10
MOUSEEVENTF_MIDDLEDOWN = 0x20
MOUSEEVENTF_MIDDLEUP = 0x40
MOUSEEVENTF_XDOWN = 0x0080
MOUSEEVENTF_XUP = 0x0100

simulated_mouse_codes = {
    (WHEEL, HORIZONTAL): MOUSEEVENTF_HWHEEL,
    (WHEEL, VERTICAL): MOUSEEVENTF_WHEEL,

    (DOWN, LEFT): MOUSEEVENTF_LEFTDOWN,
    (UP, LEFT): MOUSEEVENTF_LEFTUP,

    (DOWN, RIGHT): MOUSEEVENTF_RIGHTDOWN,
    (UP, RIGHT): MOUSEEVENTF_RIGHTUP,

    (DOWN, MIDDLE): MOUSEEVENTF_MIDDLEDOWN,
    (UP, MIDDLE): MOUSEEVENTF_MIDDLEUP,

    (DOWN, X): MOUSEEVENTF_XDOWN,
    (UP, X): MOUSEEVENTF_XUP,
}

NULL = c_int(0)

WHEEL_DELTA = 120

init = lambda: None

def listen(queue):
    def low_level_mouse_handler(nCode, wParam, lParam):
        struct = lParam.contents
        # Can't use struct.time because it's usually zero.
        t = time.time()

        if wParam == WM_MOUSEMOVE:
            event = MoveEvent(struct.x, struct.y, t)
        elif wParam == WM_MOUSEWHEEL:
            event = WheelEvent(struct.data / (WHEEL_DELTA * (2<<15)), t)
        elif wParam in buttons_by_wm_code:
            type, button = buttons_by_wm_code.get(wParam, ('?', '?'))
            if wParam >= WM_XBUTTONDOWN:
                button = {0x10000: X, 0x20000: X2}[struct.data]
            event = ButtonEvent(type, button, t)

        queue.put(event)
        return CallNextHookEx(NULL, nCode, wParam, lParam)

    WH_MOUSE_LL = c_int(14)
    mouse_callback = LowLevelMouseProc(low_level_mouse_handler)
    mouse_hook = SetWindowsHookEx(WH_MOUSE_LL, mouse_callback, NULL, NULL)

    # Register to remove the hook when the interpreter exits. Unfortunately a
    # try/finally block doesn't seem to work here.
    atexit.register(UnhookWindowsHookEx, mouse_hook)

    msg = LPMSG()
    while not GetMessage(msg, NULL, NULL, NULL):
        TranslateMessage(msg)
        DispatchMessage(msg)

def _translate_button(button):
    if button == X or button == X2:
        return X, {X: 0x10000, X2: 0x20000}[button]
    else:
        return button, 0

def press(button=LEFT):
    button, data = _translate_button(button)
    code = simulated_mouse_codes[(DOWN, button)]
    user32.mouse_event(code, 0, 0, data, 0)

def release(button=LEFT):
    button, data = _translate_button(button)
    code = simulated_mouse_codes[(UP, button)]
    user32.mouse_event(code, 0, 0, data, 0)

def wheel(delta=1):
    code = simulated_mouse_codes[(WHEEL, VERTICAL)]
    user32.mouse_event(code, 0, 0, int(delta * WHEEL_DELTA), 0)

def move_to(x, y):
    user32.SetCursorPos(int(x), int(y))

def move_relative(x, y):
    user32.mouse_event(MOUSEEVENTF_MOVE, int(x), int(y), 0, 0)

class POINT(Structure):
    _fields_ = [("x", c_long), ("y", c_long)]

def get_position():
    point = POINT()
    user32.GetCursorPos(byref(point))
    return (point.x, point.y)

if __name__ == '__main__':
    def p(e):
        print(e)
    listen(p)



import Quartz
import pygetwindow


def getAllTitles():
    """Returns a list of strings of window titles for all visible windows.
    """

    # Source: https://stackoverflow.com/questions/53237278/obtain-list-of-all-window-titles-on-macos-from-a-python-script/53985082#53985082
    windows = Quartz.CGWindowListCopyWindowInfo(Quartz.kCGWindowListExcludeDesktopElements | Quartz.kCGWindowListOptionOnScreenOnly, Quartz.kCGNullWindowID)
    return ['%s %s' % (win[Quartz.kCGWindowOwnerName], win.get(Quartz.kCGWindowName, '')) for win in windows]


def getActiveWindow():
    """Returns a Window object of the currently active Window."""

    # Source: https://stackoverflow.com/questions/5286274/front-most-window-using-cgwindowlistcopywindowinfo
    windows = Quartz.CGWindowListCopyWindowInfo(Quartz.kCGWindowListExcludeDesktopElements | Quartz.kCGWindowListOptionOnScreenOnly, Quartz.kCGNullWindowID)
    for win in windows:
        if win['kCGWindowLayer'] == 0:
            return '%s %s' % (win[Quartz.kCGWindowOwnerName], win.get(Quartz.kCGWindowName, '')) # Temporary. For now, we'll just return the title of the active window.
    raise Exception('Could not find an active window.') # Temporary hack.


def getWindowsAt(x, y):
    windows = Quartz.CGWindowListCopyWindowInfo(Quartz.kCGWindowListExcludeDesktopElements | Quartz.kCGWindowListOptionOnScreenOnly, Quartz.kCGNullWindowID)
    matches = []
    for win in windows:
        w = win['kCGWindowBounds']
        if pygetwindow.pointInRect(x, y, w['X'], w['Y'], w['Width'], w['Height']):
            matches.append('%s %s' % (win[Quartz.kCGWindowOwnerName], win.get(Quartz.kCGWindowName, '')))
    return matches



def activate():
    # TEMP - this is not a real api, I'm just using this name to store these notes for now.

    # Source: https://stackoverflow.com/questions/7460092/nswindow-makekeyandorderfront-makes-window-appear-but-not-key-or-front?rq=1
    # Source: https://stackoverflow.com/questions/4905024/is-it-possible-to-bring-window-to-front-without-taking-focus?rq=1
    pass


def getWindowGeometry(title):
    # TEMP - this is not a real api, I'm just using this name to stoe these notes for now.
    windows = Quartz.CGWindowListCopyWindowInfo(Quartz.kCGWindowListExcludeDesktopElements | Quartz.kCGWindowListOptionOnScreenOnly, Quartz.kCGNullWindowID)
    for win in windows:
        if title in '%s %s' % (win[Quartz.kCGWindowOwnerName], win.get(Quartz.kCGWindowName, '')):
            w = win['kCGWindowBounds']
            return (w['X'], w['Y'], w['Width'], w['Height'])


def isVisible(title):
    # TEMP - this is not a real api, I'm just using this name to stoe these notes for now.
    windows = Quartz.CGWindowListCopyWindowInfo(Quartz.kCGWindowListExcludeDesktopElements | Quartz.kCGWindowListOptionOnScreenOnly, Quartz.kCGNullWindowID)
    for win in windows:
        if title in '%s %s' % (win[Quartz.kCGWindowOwnerName], win.get(Quartz.kCGWindowName, '')):
            return win['kCGWindowAlpha'] != 0.0

def isMinimized():
    # TEMP - this is not a real api, I'm just using this name to stoe these notes for now.
    # Source: https://stackoverflow.com/questions/10258676/how-to-know-whether-a-window-is-minimised-or-not
    # Use the kCGWindowIsOnscreen to check this. Minimized windows are considered to not be on the screen. (But I'm not sure if there are other situations where a window is "off screen".)

    # I'm not sure how kCGWindowListOptionOnScreenOnly interferes with this.
    pass

# TODO: This class doesn't work yet. I've copied the Win32Window class and will make adjustments as needed here.

class MacOSWindow():
    def __init__(self, hWnd):
        self._hWnd = hWnd # TODO fix this, this is a LP_c_long insead of an int.

        def _onRead(attrName):
            r = self._getWindowRect(_hWnd)
            self._rect._left = r.left # Setting _left directly to skip the onRead.
            self._rect._top = r.top # Setting _top directly to skip the onRead.
            self._rect._width = r.right - r.left # Setting _width directly to skip the onRead.
            self._rect._height = r.bottom - r.top # Setting _height directly to skip the onRead.

        def _onChange(oldBox, newBox):
            self.moveTo(newBox.left, newBox.top)
            self.resizeTo(newBox.width, newBox.height)

        r = self._getWindowRect(_hWnd)
        self._rect = pyrect.Rect(r.left, r.top, r.right - r.left, r.bottom - r.top, onChange=_onChange, onRead=_onRead)

    def __str__(self):
        r = self._getWindowRect(_hWnd)
        width = r.right - r.left
        height = r.bottom - r.top
        return '<%s left="%s", top="%s", width="%s", height="%s", title="%s">' % (self.__class__.__name__, r.left, r.top, width, height, self.title)


    def __repr__(self):
        return '%s(hWnd=%s)' % (self.__class__.__name__, self._hWnd)


    def __eq__(self, other):
        return isinstance(other, MacOSWindow) and self._hWnd == other._hWnd


    def close(self):
        """Closes this window. This may trigger "Are you sure you want to
        quit?" dialogs or other actions that prevent the window from
        actually closing. This is identical to clicking the X button on the
        window."""
        raise NotImplementedError

    def minimize(self):
        """Minimizes this window."""
        raise NotImplementedError


    def maximize(self):
        """Maximizes this window."""
        raise NotImplementedError


    def restore(self):
        """If maximized or minimized, restores the window to it's normal size."""
        raise NotImplementedError


    def activate(self):
        """Activate this window and make it the foreground window."""
        raise NotImplementedError


    def resizeRel(self, widthOffset, heightOffset):
        """Resizes the window relative to its current size."""
        raise NotImplementedError


    def resizeTo(self, newWidth, newHeight):
        """Resizes the window to a new width and height."""
        raise NotImplementedError


    def moveRel(self, xOffset, yOffset):
        """Moves the window relative to its current position."""
        raise NotImplementedError


    def moveTo(self, newLeft, newTop):
        """Moves the window to new coordinates on the screen."""
        raise NotImplementedError


    @property
    def isMinimized(self):
        """Returns True if the window is currently minimized."""
        raise NotImplementedError

    @property
    def isMaximized(self):
        """Returns True if the window is currently maximized."""
        raise NotImplementedError

    @property
    def isActive(self):
        """Returns True if the window is currently the active, foreground window."""
        raise NotImplementedError

    @property
    def title(self):
        """Returns the window title as a string."""
        raise NotImplementedError

    @property
    def visible(self):
        raise NotImplementedError





# -*- coding: utf-8 -*-
import keyboard
import fileinput
import json
import sys

def print_event_json(event):
    print(event.to_json(ensure_ascii=sys.stdout.encoding != 'utf-8'))
    sys.stdout.flush()
keyboard.hook(print_event_json)

parse_event_json = lambda line: keyboard.KeyboardEvent(**json.loads(line))
keyboard.play(parse_event_json(line) for line in fileinput.input())


# !/usr/bin/python
# coding=utf-8


class Iter():
	'''
	'''
	@staticmethod
	def makeList(x):
		'''Convert the given obj to a list.

		:Parameters:
			x () = The object to convert to a list if not already a list, set, or tuple.

		:Return:
			(list)
		'''
		return list(x) if isinstance(x, (list, tuple, set, dict, range)) else [x]


	@classmethod
	def formatReturn(cls, rtn, orig=None):
		'''Return the list element if the given iterable only contains a single element.
		If the list contains multiple elements, always return the full list.
		If the 'orig' arg is a multi-element type then the original format will always be returned.

		:Parameters:
			rtn (list): An iterable.
			orig (obj): Optionally; derive the return type form the original value.
					ie. if it was a multi-value type; do not modify the return value.
		:Return:
			(obj)(list) dependant on flags.
		'''
		orig = isinstance(orig, (list, tuple, set, dict, range))

		try:
			if len(rtn)==1 and not orig and not isinstance(rtn, str):
				return rtn[0]

		except Exception as e:
			pass
		return rtn


	@classmethod
	def nestedDepth(cls, lst, typ=(list, set, tuple)):
		'''Get the maximum nested depth of any sub-lists of the given list.
		If there is nothing nested, 0 will be returned.

		:Parameters:
			lst (list): The list to check.
			typ (type)(tuple): The type(s) to include in the query.

		:Return:
			(int) 0 if none, else the max nested depth.
		'''
		d=-1
		for i in lst:
			if isinstance(i, typ):
				d = max(cls.nestedDepth(i), d)
		return d+1


	@classmethod
	def flatten(cls, lst):
		'''Flatten arbitrarily nested lists.

		:Parameters:
			lst (list): A list with potentially nested lists.

		:Return:
			(generator)
		'''
		for i in lst:
			if isinstance(i, (list, tuple, set)):
				for ii in cls.flatten(i):
					yield ii
			else:
				yield i


	@staticmethod
	def collapseList(lst, limit=None, compress=True, toString=True):
		'''Convert a list of integers to a collapsed sequential string format.
		ie. [19,22,23,24,25,26] to ['19', '22..26']

		:Parameters:
			lst (list): A list of integers.
			limit (int): limit the maximum length of the returned elements.
			compress (bool): Trim redundant chars from the second half of a compressed set. ie. ['19', '22-32', '1225-6'] from ['19', '22..32', '1225..1226']
			toString (bool): Return a single string value instead of a list.

		:Return:
			(list)(str) string if 'toString'.
		'''
		ranges=[]
		for x in map(str, lst): #make sure the list is made up of strings.
			if not ranges:
				ranges.append([x])
			elif int(x)-prev_x==1:
				ranges[-1].append(x)
			else:
				ranges.append([x])
			prev_x = int(x)

		if compress: #style: ['19', '22-32', '1225-6']
			collapsedList = ['-'.join([r[0], r[-1][len(str(r[-1]))-len(str((int(r[-1])-int(r[0])))):]] #find the difference and use that value to further trim redundant chars from the string
								if len(r) > 1 else r) 
									for r in ranges]

		else: #style: ['19', '22..32', '1225..1226']
			collapsedList = ['..'.join([r[0], r[-1]] 
								if len(r) > 1 else r) 
									for r in ranges]

		if limit and len(collapsedList)>limit:
			l = collapsedList[:limit]
			l.append('...')
			collapsedList = l
		
		if toString:
			collapsedList = ', '.join(collapsedList)

		return collapsedList


	@staticmethod
	def bitArrayToList(bitArray):
		'''Convert a binary bitArray to a python list.

		:Parameters:
			bitArray () = A bit array or list of bit arrays.

		:Return:
			(list) containing values of the indices of the on (True) bits.
		'''
		if len(bitArray):
			if type(bitArray[0])!=bool: #if list of bitArrays: flatten
				lst=[]
				for array in bitArray:
					lst.append([i+1 for i, bit in enumerate(array) if bit==1])
				return [bit for array in lst for bit in array]

			return [i+1 for i, bit in enumerate(bitArray) if bit==1]


	@staticmethod
	def rindex(itr, item):
		'''Get the index of the first item to match the given item 
		starting from the back (right side) of the list.

		:Parameters:
			itr (iter): An iterable.
			item () = The item to get the index of.

		:Return:
			(int) -1 if element not found.
		'''
		return next(iter(i for i in range(len(itr)-1,-1,-1) if itr[i]==item), -1)


	@staticmethod
	def indices(itr, value):
		'''Get the index of each element of a list matching the given value.

		:Parameters:
			itr (iter): An iterable.
			value () = The search value.

		:Return:
			(generator)
		'''
		return (i for i, v in enumerate(itr) if v==value)


	@staticmethod
	def removeDuplicates(lst, trailing=True):
		'''Remove all duplicated occurences while keeping the either the first or last.

		:Parameters:
			lst (list): The list to remove duplicate elements of.
			trailing (bool): Remove all trailing occurances while keeping the first, else keep last.

		:Return:
			(list)
		'''
		if trailing:
			return list(dict.fromkeys(lst))
		else:
			return list(dict.fromkeys(lst[::-1]))[::-1] #reverse the list when removing from the start of the list.


	@classmethod
	def filterDict(cls, dct, inc=[], exc=[], keys=False, values=False):
		'''Filter the given dictionary.
		Extends `filterList` to operate on either the given dict's keys or values.

		:Parameters:
			dct (dict): The dictionary to filter.
			inc (str)(obj)(list): The objects(s) to include.
					supports using the '*' operator: startswith*, *endswith, *contains*
					Will include all items that satisfy ANY of the given search terms.
					meaning: '*.png' and '*Normal*' returns all strings ending in '.png' AND all 
					strings containing 'Normal'. NOT strings satisfying both terms.
			exc (str)(obj)(list): The objects(s) to exclude. Similar to include.
					exlude take precidence over include.
			keys (bool): Filter the dictionary keys.
			values (bool): Filter the dictionary values.

		:Return:
			(dict)

		:Example: dct = {1:'1', 'two':2, 3:'three'}
		filterDict(dct, exc='*t*', values=True) #returns: {1: '1', 'two': 2}
		filterDict(dct, exc='t*', keys=True) #returns: {1: '1', 3: 'three'}
		filterDict(dct, exc=1, keys=True) #returns: {'two': 2, 3: 'three'}
		'''
		if keys:
			filtered = cls.filterList(dct.keys(), inc, exc)
			dct = {k: dct[k] for k in filtered}
		if values:
			filtered = cls.filterList(dct.values(), inc, exc)
			dct = {k:v for k,v in dct.items() if v in filtered}
		return dct


	@classmethod
	def filterList(cls, lst, inc=[], exc=[]):
		'''Filter the given list.

		:Parameters:
			lst (list): The components(s) to filter.
			inc (str)(int)(obj)(list): The objects(s) to include.
					supports using the '*' operator: startswith*, *endswith, *contains*
					Will include all items that satisfy ANY of the given search terms.
					meaning: '*.png' and '*Normal*' returns all strings ending in '.png' AND all 
					strings containing 'Normal'. NOT strings satisfying both terms.
			exc (str)(int)(obj)(list): The objects(s) to exclude. Similar to include.
					exlude take precidence over include.
		:Return:
			(list)

		:Example: filterList([0, 1, 2, 3, 2], [1, 2, 3], 2) #returns: [1, 3]
		'''
		exc = cls.makeList(exc)
		inc = cls.makeList(inc)

		if not any((i for i in inc+exc if isinstance(i, str) and '*' in i)): #if no wildcards used:
			return [i for i in lst if not i in exc and (i in inc if inc else i not in inc)]

		#else: split `inc` and `exc` lists into separate tuples according to wildcard positions. 
		if exc:
			exc_, excContains, excStartsWith, excEndsWith = [],[],[],[]
			for i in exc:
				if isinstance(i, str) and '*' in i:
					if i.startswith('*'):
						if i.endswith('*'):
							excContains.append(i[1:-1])
						excEndsWith.append(i[1:])
					elif i.endswith('*'):
						excStartsWith.append(i[:-1])
				else:
					exc_.append(i)
		if inc:
			inc_, incContains, incStartsWith, incEndsWith = [],[],[],[]
			for i in inc:
				if isinstance(i, str) and '*' in i:
					if i.startswith('*'):
						if i.endswith('*'):
							incContains.append(i[1:-1])
						incEndsWith.append(i[1:])
					elif i.endswith('*'):
						incStartsWith.append(i[:-1])
				else:
					inc_.append(i)

		result=[]
		for i in lst:

			if exc:
				if i in exc_ or isinstance(i, str) and any((
					i.startswith(tuple(excStartsWith)),
					i.endswith(tuple(excEndsWith)),
					next(iter(chars in i for chars in excContains), False))):
					continue

			if inc:
				if i not in inc_ and not (isinstance(i, str) and any(( 
					i.startswith(tuple(incStartsWith)), 
					i.endswith(tuple(incEndsWith)), 
					next(iter(chars in i for chars in incContains), False)))):
					continue

			result.append(i)
		return result


	@staticmethod
	def splitList(lst, into):
		'''Split a list into parts.

		:Parameters:
			into (str): Split the list into parts defined by the following:
				'<n>parts' - Split the list into n parts.
					ex. 2 returns:  [[1, 2, 3, 5], [7, 8, 9]] from [1,2,3,5,7,8,9]
				'<n>parts+' - Split the list into n equal parts with any trailing remainder.
					ex. 2 returns:  [[1, 2, 3], [5, 7, 8], [9]] from [1,2,3,5,7,8,9]
				'<n>chunks' - Split into sublists of n size.
					ex. 2 returns: [[1,2], [3,5], [7,8], [9]] from [1,2,3,5,7,8,9]
				'contiguous' - The list will be split by contiguous numerical values.
					ex. 'contiguous' returns: [[1,2,3], [5], [7,8,9]] from [1,2,3,5,7,8,9]
				'range' - The values of 'contiguous' will be limited to the high and low end of each range.
					ex. 'range' returns: [[1,3], [5], [7,9]] from [1,2,3,5,7,8,9]
		:Return:
			(list)
		'''
		from string import digits, ascii_letters, punctuation
		mode = into.lower().lstrip(digits)
		digit = into.strip(ascii_letters+punctuation)
		n = int(digit) if digit else None

		if n:
			if mode=='parts':
				n = len(lst)*-1 // n*-1 #ceil
			elif mode=='parts+':
				n = len(lst) // n
			return [lst[i:i+n] for i in range(0, len(lst), n)]

		elif mode=='contiguous' or mode=='range':
			from itertools import groupby
			from operator import itemgetter

			try:
				contiguous = [list(map(itemgetter(1), g)) for k, g in groupby(enumerate(lst), lambda x: int(x[0])-int(x[1]))]
			except ValueError as error:
				print ('{} in splitList\n\t# Error: {} #\n\t{}'.format(__file__, error, lst))
				return lst
			if mode=='range':
				return [[i[0], i[-1]] if len(i)>1 else (i) for i in contiguous]
			return contiguous

# --------------------------------------------------------------------------------------------









# --------------------------------------------------------------------------------------------

def __getattr__(attr:str):
	"""Searches for an attribute in this module's classes and returns it.

	:Parameters:
		attr (str): The name of the attribute to search for.
	
	:Return:
		(obj) The found attribute.

	:Raises:
		AttributeError: If the given attribute is not found in any of the classes in the module.
	"""
	try:
		return getattr(Iter, attr)

	except AttributeError as error:
		raise AttributeError(f"Module '{__name__}' has no attribute '{attr}'")

# --------------------------------------------------------------------------------------------

if __name__=='__main__':
	pass

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# Deprecated ------------------------------------